"""Constants - Application global constants."""

from typing import Final

APP_NAME: Final = "Tony"
VERSION: Final = "1.6.1"
PACKAGE_NAME: Final = "neuro-api-tony"
GIT_REPO_URL: Final = "https://github.com/Pasu4/neuro-api-tony"
PYPI_URL: Final = "https://pypi.org/project/neuro-api-tony"
PYPI_API_URL: Final = "https://pypi.org/pypi/neuro-api-tony/json"



"""Controller - Application controller module."""

from __future__ import annotations

import json
import random
from typing import TYPE_CHECKING, Any

import wx
from jsf import JSF

from .api import (
    ActionResultCommand,
    ActionsForceCommand,
    ActionsRegisterCommand,
    ActionsUnregisterCommand,
    ContextCommand,
    NeuroAPI,
    ShutdownReadyCommand,
    StartupCommand,
)
from .constants import VERSION
from .model import NeuroAction, TonyModel
from .view import TonyView

if TYPE_CHECKING:
    from collections.abc import Generator


def action_id_generator() -> Generator[str, None, None]:
    """Generate a unique ID for an action."""
    i = 0
    while True:
        yield f"action_{i}"
        i += 1


class TonyController:
    """TonyController class."""

    def __init__(self, app: wx.App, log_level: str) -> None:
        """Initialize Tony Controller."""
        self.app = app
        self.model = TonyModel()
        self.api = NeuroAPI(wx.CallAfter)
        self.view = TonyView(app, self.model, log_level, self.api.on_close)

        self.active_actions_force: ActionsForceCommand | None = None

        self.id_generator = action_id_generator()

        self.inject()

    def run(self, address: str, port: int, init_message: str) -> None:
        """Start websocket server on given address and run GUI main event loop."""
        # Schedule the API start to run after the main loop starts
        wx.CallAfter(self.api.start, address, port)
        wx.CallAfter(self.view.log_info, f"Running version {VERSION}")
        if init_message:
            wx.CallAfter(self.view.log_info, init_message)

        self.view.show()
        self.app.MainLoop()

    def inject(self) -> None:
        """Inject methods into the view and API."""
        # fmt: off
        self.api.on_startup = self.on_startup
        self.api.on_context = self.on_context
        self.api.on_actions_register = self.on_actions_register
        self.api.on_actions_unregister = self.on_actions_unregister
        self.api.on_actions_force = self.on_actions_force
        self.api.on_action_result = self.on_action_result
        self.api.on_shutdown_ready = self.on_shutdown_ready
        self.api.on_unknown_command = self.on_unknown_command
        self.api.log_command = self.view.log_command
        self.api.log_debug = self.view.log_debug
        self.api.log_info = self.view.log_info
        self.api.log_warning = self.view.log_warning
        self.api.log_error = self.view.log_error
        self.api.log_critical = self.view.log_critical
        self.api.log_raw = self.view.log_raw
        self.api.get_delay = lambda: float(self.view.controls.latency / 1000)

        self.view.on_execute = self.on_view_execute
        self.view.on_delete_action = self.on_view_delete_action
        self.view.on_delete_all_actions = self.on_view_delete_all_actions
        self.view.on_unlock = self.on_view_unlock
        self.view.on_clear_logs = self.on_view_clear_logs
        self.view.on_send_actions_reregister_all = self.on_view_send_actions_reregister_all
        self.view.on_send_shutdown_graceful = self.on_view_send_shutdown_graceful
        self.view.on_send_shutdown_graceful_cancel = self.on_view_send_shutdown_graceful_cancel
        self.view.on_send_shutdown_immediate = self.on_view_send_shutdown_immediate
        # fmt: on

    def on_any_command(self, cmd: Any) -> None:
        """Handle any command received from the API."""

    def on_startup(self, cmd: StartupCommand) -> None:
        """Handle the startup command."""
        # TODO: Change to trigger on client connection instead of startup message
        self.view.log_info(f'Started game "{cmd.game}"')

        self.model.clear_actions()
        self.view.clear_actions()

    def on_context(self, cmd: ContextCommand) -> None:
        """Handle the context command."""
        self.view.log_context(cmd.message, silent=cmd.silent)

    def on_actions_register(self, cmd: ActionsRegisterCommand) -> None:
        """Handle the actions/register command."""
        for action in cmd.actions:
            # Check if an action with the same name already exists
            if self.model.has_action(action.name):
                self.view.log_warning(f'Action "{action.name}" already exists. Ignoring.')
                continue

            self.model.add_action(action)
            wx.CallAfter(self.view.add_action, action)
            self.view.log_description(f"{action.name}: {action.description}")
        s = "s" if len(cmd.actions) != 1 else ""
        self.view.log_info(f"Action{s} registered: {', '.join(action.name for action in cmd.actions)}")

    def on_actions_unregister(self, cmd: ActionsUnregisterCommand) -> None:
        """Handle the actions/unregister command."""
        known_actions = [name for name in cmd.action_names if self.model.has_action(name)]
        unknown_actions = [name for name in cmd.action_names if not self.model.has_action(name)]
        for name in known_actions:
            self.model.remove_action_by_name(name)
            self.view.remove_action_by_name(name)
        s1 = "s" if len(cmd.action_names) != 1 else ""
        s2 = "s" if len(unknown_actions) != 1 else ""
        if known_actions:
            self.view.log_info(f"Action{s1} unregistered: {', '.join(known_actions)}")
        if unknown_actions:
            self.view.log_info(f"Ignoring unregistration of unknown action{s2}: {', '.join(unknown_actions)}")
        if not known_actions and not unknown_actions:
            self.view.log_warning("No actions to unregister specified.")

    def on_actions_force(self, cmd: ActionsForceCommand) -> None:
        """Handle the actions/force command."""
        if cmd.state is not None and cmd.state != "":
            self.view.log_state(cmd.state, cmd.ephemeral_context)
        else:
            self.view.log_info("actions/force command contains no state.")

        self.view.log_query(cmd.query, cmd.ephemeral_context)

        if self.view.controls.ignore_actions_force:
            self.view.log_info("Forced action ignored.")  # TODO: Make this configurable as warning?
            self.active_actions_force = None
            return

        # Check if all actions exist
        if not all(self.model.has_action(name) for name in cmd.action_names):
            self.view.log_warning(
                "actions/force with invalid actions received. Discarding.\nInvalid actions: "
                + ", ".join(name for name in cmd.action_names if not self.model.has_action(name)),
            )
            self.active_actions_force = None
            return

        self.execute_actions_force(cmd)

    def on_action_result(self, cmd: ActionResultCommand) -> None:
        """Handle the action/result command."""
        self.view.log_info("Action result indicates " + ("success" if cmd.success else "failure"))

        self.view.log_debug(f"cmd.success: {cmd.success}, active_actions_force: {self.active_actions_force}")

        if not cmd.success and self.active_actions_force is not None:
            self.retry_actions_force(self.active_actions_force)
        else:
            self.active_actions_force = None

        if cmd.message is not None:
            self.view.log_action_result(cmd.success, cmd.message)
        elif cmd.success:
            self.view.log_info("Successful action result contains no message.")
        else:
            self.view.log_warning("Failed action result contains no message.")

        wx.CallAfter(self.view.on_action_result, cmd.success, cmd.message)

    def on_shutdown_ready(self, cmd: ShutdownReadyCommand) -> None:
        """Handle the shutdown/ready command."""
        self.view.log_info("shutdown/ready is not officially supported.")

    def on_unknown_command(self, json_cmd: Any) -> None:
        """Handle an unknown command."""

        # self.view.log_warning(f'Unknown command received: {json_cmd['command']}')

    def send_action(self, id_: str, name: str, data: str | None) -> None:
        """Send an action command to the API."""
        self.view.log_info(f"Sending action: {name}")
        self.api.send_action(id_, name, data)

        # Disable the actions until the result is received
        self.view.disable_actions()

    def send_actions_reregister_all(self) -> None:
        """Send an actions/reregister_all command to the API."""
        self.api.send_actions_reregister_all()

    def on_view_execute(self, action: NeuroAction) -> bool:
        """Handle an action execution request from the view.

        Returns True if an action was sent, False if the action was cancelled.
        """
        if not action.schema:
            self.send_action(
                next(self.id_generator),
                action.name,
                None,
            )  # No schema, so send the action immediately
            return True

        # If there is a schema, open a dialog to get the data
        result = self.view.show_action_dialog(action)
        if result is None:
            return False  # User cancelled the dialog

        self.model.last_action_data[action.name] = result  # Store the last data in the action object
        self.send_action(next(self.id_generator), action.name, result)
        return True

    def on_view_delete_action(self, name: str) -> None:
        """Handle a request to delete an action from the view."""
        self.model.remove_action_by_name(name)
        self.view.remove_action_by_name(name)

        self.view.log_info(f"Action deleted: {name}")

    def on_view_delete_all_actions(self) -> None:
        """Handle a request to delete all actions from the view."""
        self.model.clear_actions()
        self.view.clear_actions()
        self.view.log_info("All actions deleted.")

    def on_view_unlock(self) -> None:
        """Handle a request to unlock the view."""
        self.view.log_info("Stopped waiting for action result.")  # TODO: Make this configurable as warning?
        self.view.enable_actions()

    def on_view_clear_logs(self) -> None:
        """Handle a request to clear the logs from the view."""
        self.view.clear_logs()
        self.view.log_info("Logs cleared.")

    def on_view_send_actions_reregister_all(self) -> None:
        """Handle a request to send an actions/reregister_all command from the view."""
        self.model.clear_actions()
        wx.CallAfter(self.view.clear_actions)
        self.send_actions_reregister_all()

    def on_view_send_shutdown_graceful(self) -> None:
        """Handle a request to send a shutdown/graceful command with wants_shutdown=true from the view."""
        self.api.send_shutdown_graceful(True)

    def on_view_send_shutdown_graceful_cancel(self) -> None:
        """Handle a request to send a shutdown/graceful with wants_shutdown=false command from the view."""
        self.api.send_shutdown_graceful(False)

    def on_view_send_shutdown_immediate(self) -> None:
        """Handle a request to send a shutdown/immediate command from the view."""
        self.api.send_shutdown_immediate()

    def execute_actions_force(
        self,
        cmd: ActionsForceCommand,
        retry: bool = False,
    ) -> None:
        """Handle a request from the game to execute a forced action."""
        self.active_actions_force = cmd

        if self.view.controls.auto_send:
            self.view.log_info("Automatically sending random action.")
            actions = [action for action in self.model.actions if action.name in cmd.action_names]
            # S311 - Standard pseudo-random generators are not suitable for cryptographic purposes
            # Not using for cryptographic purposes so we should be fine
            action = random.choice(actions)  # noqa: S311

            if not action.schema:
                self.send_action(next(self.id_generator), action.name, None)
            else:
                faker = JSF(action.schema)
                sample = faker.generate()
                self.send_action(
                    next(self.id_generator),
                    action.name,
                    json.dumps(sample),
                )

        else:
            wx.CallAfter(
                self.view.force_actions,
                cmd.state,
                cmd.query,
                cmd.ephemeral_context,
                cmd.action_names,
                retry,
            )

    def retry_actions_force(self, cmd: ActionsForceCommand) -> None:
        """Retry the actions/force command."""
        if self.view.controls.ignore_actions_force:
            self.view.log_warning("Forced action ignored.")
            self.active_actions_force = None
            return

        # Check if all actions exist
        if not all(self.model.has_action(name) for name in cmd.action_names):
            self.view.log_warning(
                "Actions have been unregistered before retrying the forced action. Retry aborted.\nInvalid actions: "
                + ", ".join(name for name in cmd.action_names if not self.model.has_action(name)),
            )
            self.active_actions_force = None
            return

        self.view.log_info("Retrying forced action.")

        self.execute_actions_force(cmd, retry=True)



from __future__ import annotations

import sys
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from json import JSONDecodeError
from typing import TYPE_CHECKING, Any
from unittest.mock import AsyncMock, MagicMock, Mock, patch

import pytest
import trio
from exceptiongroup import catch
from trio_websocket import WebSocketRequest

if sys.version_info < (3, 11):
    from exceptiongroup import BaseExceptionGroup

from neuro_api_tony.api import ActionsRegisterCommand, NeuroAPI
from neuro_api_tony.model import NeuroAction

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable, Generator


@pytest.fixture
def api() -> NeuroAPI:
    """Create a NeuroAPI instance for testing."""
    return NeuroAPI(Mock())


@pytest.fixture(autouse=True)
def print_exception_to_raise() -> Generator[None, None, None]:
    """Make traceback.print_exception raise exception instead of printing it."""

    def raise_(exc: BaseException) -> None:
        raise exc from None

    with patch(
        "traceback.print_exception",
        raise_,
    ):
        yield None


def test_start(api: NeuroAPI) -> None:
    """Test starting the WebSocket server."""
    with patch(
        "trio.lowlevel.start_guest_run",
        new_callable=Mock,
    ) as mock_start:
        api.start("localhost", 8080)
        assert api._async_library_running

        # Attempt to start again
        api.start("localhost", 8080)

        # Ensure start only called once
        assert mock_start.call_count == 1


def test_run_start_stop(api: NeuroAPI) -> None:
    """Test starting and stopping the WebSocket server."""
    tasks: list[Callable[[], None]] = []

    def run_sync_soon_threadsafe(func: Callable[[], None]) -> None:
        tasks.append(func)

    def run_tasks_until_empty() -> None:
        # Limit to 100 iterations
        for _ in range(100):
            if not tasks:
                break
            tasks.pop(0)()

    with patch.object(
        api,
        "run_sync_soon_threadsafe",
        run_sync_soon_threadsafe,
    ):
        api.start("localhost", 8080)

    assert api._async_library_running
    api.stop()

    # 2nd cancel shouldn't lead to any problems
    api.stop()

    assert api._async_library_running

    close_mock = Mock()
    with patch.object(
        api,
        "run_sync_soon_threadsafe",
        run_sync_soon_threadsafe,
    ):
        api.on_close(close_mock)

        # 2nd start should be ignored while shutting down
        api.start("localhost", 8080)

        # 2nd close handle should be skipped
        api.on_close(close_mock)

        run_tasks_until_empty()

    close_mock.assert_called_once()

    assert not api._async_library_running
    # Make sure stop when not running is also fine
    api.stop()  # type: ignore[unreachable]


def test_run_start_failure(api: NeuroAPI) -> None:
    """Test that async_library_running is handled even if guest run start fails."""

    def start_guest_run(*args: Any, **kwargs: Any) -> None:
        raise ValueError("jerald")

    with pytest.raises(ValueError, match="^jerald$"):
        with patch(
            "trio.lowlevel.start_guest_run",
            start_guest_run,
        ):
            api.start("localhost", 8080)
    assert not api._async_library_running


@pytest.mark.trio
async def test_handle_websocket_request_reject(api: NeuroAPI) -> None:
    """Test rejecting a WebSocket connection request when already connected."""
    send, receive = trio.open_memory_channel[str](0)
    api._message_send_channel = send
    mock_request = MagicMock()
    mock_request.reject = AsyncMock()

    await api._handle_websocket_request(mock_request)

    mock_request.reject.assert_called_once_with(
        503,
        body=b"Server does not support multiple connections at once currently",
    )


def test_send_action(api: NeuroAPI) -> None:
    """Test sending an action command."""
    send, receive = trio.open_memory_channel[str](1)
    api._message_send_channel = send
    assert api.send_action("123", "test_action", None)

    assert api._current_action_id == "123"
    assert receive.receive_nowait() == '{"command": "action", "data": {"id": "123", "name": "test_action"}}'  # fmt: skip


def test_send_action_with_data(api: NeuroAPI) -> None:
    """Test sending an action command."""
    send, receive = trio.open_memory_channel[str](1)
    api._message_send_channel = send
    assert api.send_action("123", "test_action", "data field")

    assert api._current_action_id == "123"
    assert (
        receive.receive_nowait()
        == '{"command": "action", "data": {"id": "123", "name": "test_action", "data": "data field"}}'
    )


def test_send_actions_reregister_all(api: NeuroAPI) -> None:
    """Test sending reregister all command."""
    send, receive = trio.open_memory_channel[str](1)
    api._message_send_channel = send
    assert api.send_actions_reregister_all()

    assert receive.receive_nowait() == '{"command": "actions/reregister_all"}'


def test_send_actions_reregister_all_not_connected(api: NeuroAPI) -> None:
    """Test sending reregister all command but not connected."""
    assert not api.send_actions_reregister_all()


def test_send_shutdown_graceful(api: NeuroAPI) -> None:
    """Test sending graceful shutdown command."""
    send, receive = trio.open_memory_channel[str](1)
    api._message_send_channel = send
    assert api.send_shutdown_graceful(True)

    assert receive.receive_nowait() == '{"command": "shutdown/graceful", "data": {"wants_shutdown": true}}'


def test_send_shutdown_graceful_not_connected(api: NeuroAPI) -> None:
    """Test sending graceful shutdown command."""
    assert not api.send_shutdown_graceful(True)


def test_send_shutdown_immediate(api: NeuroAPI) -> None:
    """Test sending immediate shutdown command."""
    send, receive = trio.open_memory_channel[str](1)
    api._message_send_channel = send
    assert api.send_shutdown_immediate()

    assert receive.receive_nowait() == '{"command": "shutdown/immediate"}'


def test_send_shutdown_immediate_not_connected(api: NeuroAPI) -> None:
    """Test sending immediate shutdown command but not connected."""
    assert not api.send_shutdown_immediate()


def test_send_action_no_client(api: NeuroAPI) -> None:
    """Test sending an action command when no client is connected."""
    assert not api.send_action("123", "test_action", None)

    assert api._current_action_id is None


def test_check_invalid_keys_recursive(api: NeuroAPI) -> None:
    """Test checking for invalid keys in a schema."""
    schema = {
        "valid_key": {},
        "allOf": {},
        "another_key": {
            "$vocabulary": {},
            "3rd level": [
                {
                    "additionalProperties": "seven",
                    "uses_waffle_iron": True,
                },
                "spaghetti",
            ],
        },
    }
    invalid_keys = api.check_invalid_keys_recursive(schema)

    assert invalid_keys == ["allOf", "$vocabulary", "additionalProperties"]


def test_check_invalid_keys_recursive_unhandled(api: NeuroAPI) -> None:
    """Test checking for invalid keys in a schema."""
    schema = {
        "valid_key": set(),
        "writeOnly": True,
    }
    api.log_error = Mock()

    invalid = api.check_invalid_keys_recursive(schema)
    assert invalid == ["writeOnly"]
    api.log_error.assert_called_once_with("Unhandled schema value type <class 'set'> (set())")


def test_actions_register_command() -> None:
    actions: list[dict[str, Any]] = [
        {
            "name": "jerald",
            "description": "jerald action",
        },
        {
            "name": "jerald_schema",
            "description": "jerald action with schema",
            "schema": {},
        },
    ]

    command = ActionsRegisterCommand(actions)
    assert command.actions == [
        NeuroAction("jerald", "jerald action", None),
        NeuroAction("jerald_schema", "jerald action with schema", {}),
    ]


@pytest.mark.trio
async def test_handle_websocket_request_accept(api: NeuroAPI) -> None:
    """Test handling a WebSocket connection request."""
    send, receive = trio.open_memory_channel[str](1)

    class Websocket:
        get_message = receive.receive
        send_message = send.send

    class Request(WebSocketRequest):
        def __init__(self) -> None:
            pass

        async def accept(  # type: ignore[override]
            self,
            *,
            subprotocol: str | None = None,
            extra_headers: list[tuple[bytes, bytes]] | None = None,
        ) -> AbstractAsyncContextManager[Websocket]:
            @asynccontextmanager
            async def manager() -> AsyncGenerator[Websocket, None]:
                yield Websocket()

            return manager()

    request = Request()

    async with trio.open_nursery() as nursery:
        nursery.start_soon(api._handle_websocket_request, request)
        await trio.sleep(0.05)
        nursery.cancel_scope.cancel()

    assert api._message_send_channel is None


@pytest.mark.trio
async def test_handle_client_connection(api: NeuroAPI) -> None:
    """Test handling a client connection."""
    mock_websocket = MagicMock()
    send, receive = trio.open_memory_channel[str](1)
    ws_send, ws_receive = trio.open_memory_channel[str](1)
    mock_websocket.get_message = ws_receive.receive
    mock_websocket.send_message = ws_send.send
    api._message_send_channel = send

    await send.send('{"command": "startup", "game": "test_game"}')

    async with trio.open_nursery() as nursery:
        nursery.start_soon(
            api._handle_client_connection,
            mock_websocket,
            receive,
        )
        await trio.sleep(0.05)
        nursery.cancel_scope.cancel()


@pytest.mark.trio
async def test_handle_consumer(api: NeuroAPI) -> None:
    """Test handling a consumer message."""
    mock_websocket = MagicMock()
    send, receive = trio.open_memory_channel[str](1)
    mock_websocket.get_message = receive.receive
    api._message_send_channel = send

    await send.send('{"command": "startup", "game": "test_game"}')

    async with trio.open_nursery() as nursery:
        cancel_scope = trio.CancelScope()
        nursery.start_soon(api._handle_consumer, mock_websocket, cancel_scope)
        await trio.sleep(0.05)
        cancel_scope.cancel()
        nursery.cancel_scope.cancel()

    assert api._current_game == "test_game"


@pytest.mark.trio
async def test_handle_consumer_invalid_json(api: NeuroAPI) -> None:
    """Test handling an invalid JSON message."""
    mock_websocket = MagicMock()
    send, receive = trio.open_memory_channel[str](1)
    mock_websocket.get_message = receive.receive
    api._message_send_channel = send

    # Missing closing brace
    await send.send('{"command": "startup", "game": "test_game"')

    had_json_error = False

    def handle_json_error(
        multi_exc: BaseExceptionGroup[JSONDecodeError],
    ) -> None:
        nonlocal had_json_error
        exc = multi_exc.args[1][0]
        assert exc.args[0] == "Expecting ',' delimiter: line 1 column 43 (char 42)"
        had_json_error = True

    with catch(
        {
            JSONDecodeError: handle_json_error,
        },
    ):
        async with trio.open_nursery() as nursery:
            cancel_scope = trio.CancelScope()
            nursery.start_soon(
                api._handle_consumer,
                mock_websocket,
                cancel_scope,
            )
            await trio.sleep(0.05)
            cancel_scope.cancel()
            nursery.cancel_scope.cancel()

    if not had_json_error:
        raise ValueError("Should have gotten JSONDecodeError")

    assert api._current_game == ""


@pytest.mark.trio
async def test_handle_producer(api: NeuroAPI) -> None:
    """Test handling a producer message."""
    mock_websocket = MagicMock()
    send, receive = trio.open_memory_channel[str](1)
    mock_websocket.send_message = AsyncMock(return_value=None)
    api._message_send_channel = send

    send.send_nowait('{"command": "action", "data": {"id": "123", "name": "test_action"}}')

    async with trio.open_nursery() as nursery:
        nursery.start_soon(api._handle_producer, mock_websocket, receive)
        await trio.sleep(0.05)
        nursery.cancel_scope.cancel()

    mock_websocket.send_message.assert_called_once_with(
        '{"command": "action", "data": {"id": "123", "name": "test_action"}}',
    )


@pytest.mark.trio
async def test_handle_producer_no_client(api: NeuroAPI) -> None:
    """Test handling a producer message when no client is connected."""
    send, receive = trio.open_memory_channel[str](1)

    async with trio.open_nursery() as nursery:
        nursery.start_soon(api._handle_producer, MagicMock(), receive)
        await trio.sleep(0.05)
        nursery.cancel_scope.cancel()


@pytest.mark.trio
async def test_handle_consumer_unexpected_command(api: NeuroAPI) -> None:
    """Test handling an unexpected command in the consumer."""
    mock_websocket = MagicMock()
    send, receive = trio.open_memory_channel[str](1)
    mock_websocket.get_message = receive.receive
    api._message_send_channel = send

    await send.send('{"command": "unknown_command"}')

    api.log_warning = Mock()

    async with trio.open_nursery() as nursery:
        cancel_scope = trio.CancelScope()
        nursery.start_soon(api._handle_consumer, mock_websocket, cancel_scope)
        await trio.sleep(0.05)
        cancel_scope.cancel()
        nursery.cancel_scope.cancel()

    api.log_warning.assert_called_with("Unknown command: unknown_command")


@pytest.mark.trio
async def test_handle_consumer_action_result(api: NeuroAPI) -> None:
    """Test handling an action/result command in the consumer."""
    mock_websocket = MagicMock()
    send, receive = trio.open_memory_channel[str](1)
    mock_websocket.get_message = receive.receive
    api._message_send_channel = send

    await send.send('{"command": "action/result", "data": {"id": "123", "success": true}}')

    async with trio.open_nursery() as nursery:
        cancel_scope = trio.CancelScope()
        nursery.start_soon(api._handle_consumer, mock_websocket, cancel_scope)
        await trio.sleep(0.05)
        cancel_scope.cancel()
        nursery.cancel_scope.cancel()

    assert api._current_action_id is None



import pytest

from neuro_api_tony.model import NeuroAction, TonyModel


@pytest.fixture
def model() -> TonyModel:
    """Fixture to create a TonyModel instance for testing."""
    return TonyModel()


def test_repr(model: TonyModel) -> None:
    """Test that repr works properly."""
    assert repr(model) == "TonyModel()"


def test_add_action(model: TonyModel) -> None:
    """Test adding an action to the model."""
    action = NeuroAction(
        name="test_action",
        description="A test action",
        schema=None,
    )
    model.add_action(action)
    assert model.actions == [action]


def test_remove_action(model: TonyModel) -> None:
    """Test removing an action from the model."""
    action = NeuroAction(
        name="test_action",
        description="A test action",
        schema=None,
    )
    model.add_action(action)
    model.remove_action(action)
    assert not model.actions


def test_remove_action_by_name(model: TonyModel) -> None:
    """Test removing an action by name."""
    action1 = NeuroAction(
        name="action1",
        description="First action",
        schema=None,
    )
    action2 = NeuroAction(
        name="action2",
        description="Second action",
        schema=None,
    )
    model.add_action(action1)
    model.add_action(action2)
    model.remove_action_by_name("action1")
    assert model.actions == [action2]


def test_clear_actions(model: TonyModel) -> None:
    """Test clearing all actions from the model."""
    action = NeuroAction(
        name="test_action",
        description="A test action",
        schema=None,
    )
    model.add_action(action)
    model.clear_actions()
    assert not model.actions


def test_has_action(model: TonyModel) -> None:
    """Test checking if an action exists in the model."""
    action = NeuroAction(
        name="test_action",
        description="A test action",
        schema=None,
    )
    model.add_action(action)
    assert model.has_action("test_action")
    assert not model.has_action("non_existent_action")


def test_get_action_by_name(model: TonyModel) -> None:
    """Test getting an action by name."""
    action = NeuroAction(
        name="test_action",
        description="A test action",
        schema=None,
    )
    model.add_action(action)
    retrieved_action = model.get_action_by_name("test_action")
    assert retrieved_action is action
    assert model.get_action_by_name("non_existent_action") is None



"""View - Main GUI frame logic."""

from __future__ import annotations

import json
from datetime import datetime as dt
from typing import TYPE_CHECKING

import jsonschema
import wx
from jsf import JSF

from .constants import VERSION

if TYPE_CHECKING:
    from collections.abc import Callable

    from .model import NeuroAction, TonyModel


# region Events

EVTTYPE_ADD_ACTION = wx.NewEventType()
EVT_ADD_ACTION = wx.PyEventBinder(EVTTYPE_ADD_ACTION, 1)


class AddActionEvent(wx.PyCommandEvent):  # type: ignore[misc]
    """An event for adding an action to the list."""

    __slots__ = ("action",)

    def __init__(self, id_: int, action: NeuroAction) -> None:
        """Initialize AddActionEvent."""
        super().__init__(EVTTYPE_ADD_ACTION, id_)
        self.action = action


EVTTYPE_ACTION_RESULT = wx.NewEventType()
EVT_ACTION_RESULT = wx.PyEventBinder(EVTTYPE_ACTION_RESULT, 1)


class ActionResultEvent(wx.PyCommandEvent):  # type: ignore[misc]
    """An event for an action result message."""

    __slots__ = ("message", "success")

    def __init__(self, id_: int, success: bool, message: str | None) -> None:
        """Initialize ActionResultEvent."""
        super().__init__(EVTTYPE_ACTION_RESULT, id_)
        self.success = success
        self.message = message


EVT_TYPE_EXECUTE = wx.NewEventType()
EVT_EXECUTE = wx.PyEventBinder(EVT_TYPE_EXECUTE, 1)


class ExecuteEvent(wx.PyCommandEvent):  # type: ignore[misc]
    """An event for executing an action."""

    __slots__ = ("action",)

    def __init__(self, id_: int, action: NeuroAction) -> None:
        """Initialize ExecuteEvent."""
        super().__init__(EVT_TYPE_EXECUTE, id_)
        self.action = action


# endregion

# region Constants

# Colors
# fmt: off
LOG_COLOR_DEFAULT                       = wx.Colour(  0,   0,   0)
LOG_COLOR_TIMESTAMP                     = wx.Colour(  0, 128,   0)
LOG_COLOR_DEBUG                         = wx.Colour(128, 128, 128)
LOG_COLOR_INFO                          = wx.Colour(128, 192, 255)
LOG_COLOR_WARNING                       = wx.Colour(255, 192,   0)
LOG_COLOR_ERROR                         = wx.Colour(255,   0,   0)
LOG_COLOR_CRITICAL                      = wx.Colour(192,   0,   0)
LOG_COLOR_CONTEXT                       = LOG_COLOR_DEFAULT
LOG_COLOR_CONTEXT_QUERY                 = wx.Colour(255,   0, 255)
LOG_COLOR_CONTEXT_STATE                 = wx.Colour(128, 255, 128)
LOG_COLOR_CONTEXT_SILENT                = wx.Colour(128, 128, 128)
LOG_COLOR_CONTEXT_EPHEMERAL             = wx.Colour(128, 192, 255)
LOG_COLOR_CONTEXT_ACTION                = wx.Colour(  0,   0, 255)
LOG_COLOR_CONTEXT_ACTION_RESULT_SUCCESS = wx.Colour(  0, 128,   0)
LOG_COLOR_CONTEXT_ACTION_RESULT_FAILURE = wx.Colour(255,   0,   0)
LOG_COLOR_INCOMING                      = wx.Colour(  0,   0, 255)
LOG_COLOR_OUTGOING                      = wx.Colour(255,   0, 128)
LOG_COLOR_COMMAND_ADDITION              = wx.Colour(128, 128, 128)
# fmt: on

UI_COLOR_WARNING = wx.Colour(255, 255, 128)
UI_COLOR_ERROR = wx.Colour(255, 192, 192)

LOG_LEVELS = {
    "DEBUG": 10,
    "INFO": 20,
    "WARNING": 30,
    "ERROR": 40,
    "CRITICAL": 50,
    "SYSTEM": 60,
}

LATENCY_TOOLTIP = (
    "Latency in milliseconds to add to each outgoing command."
    " Must be non-negative and not exceed 10000 ms."
)  # fmt: skip

# endregion


class TonyView:
    """The view class for Tony."""

    def __init__(
        self,
        app: wx.App,
        model: TonyModel,
        log_level: str,
        api_close: Callable[[Callable[[], None]], None],
    ) -> None:
        """Initialize TonyView."""
        self.model = model

        self.controls = Controls()
        self.controls.set_log_level(log_level)

        self.frame = MainFrame(self)
        app.SetTopWindow(self.frame)

        self.api_close = api_close
        self.frame.Bind(wx.EVT_CLOSE, self.on_close)

        self.action_dialog: ActionDialog | None = None

        # Dependency injection
        # fmt: off
        self.on_execute: Callable[[NeuroAction], bool] = lambda action: False
        self.on_delete_action: Callable[[str], None] = lambda name: None
        self.on_delete_all_actions: Callable[[], None] = lambda: None
        self.on_unlock: Callable[[], None] = lambda: None
        self.on_clear_logs: Callable[[], None] = lambda: None
        self.on_send_actions_reregister_all: Callable[[], None] = lambda: None
        self.on_send_shutdown_graceful: Callable[[], None] = lambda: None
        self.on_send_shutdown_graceful_cancel: Callable[[], None] = lambda: None
        self.on_send_shutdown_immediate: Callable[[], None] = lambda: None
        # fmt: on

    def on_close(self, event: wx.CloseEvent) -> None:
        """Handle application close event."""
        # Do not let application exit
        event.Veto()
        # Tell api to close async run cleanly and then call destroy this frame
        self.api_close(self.frame.Destroy)

    def show(self) -> None:
        """Show the main frame."""
        self.frame.Show()

    def log_command(self, command: str, incoming: bool, addition: str | None = None) -> None:
        """Log a command."""
        tag = "Game --> Tony" if incoming else "Game <-- Tony"
        color = LOG_COLOR_INCOMING if incoming else LOG_COLOR_OUTGOING

        if addition is None:
            self.add_export_log(command, tag, "Commands")
            self.frame.panel.log_notebook.command_log_panel.log(command, tag, color)
        else:
            self.add_export_log(f"{command}: {addition}", tag, "Commands")
            self.frame.panel.log_notebook.command_log_panel.log(
                [(command + ": ", LOG_COLOR_DEFAULT), (addition, LOG_COLOR_COMMAND_ADDITION)],
                tag,
                color,
            )

    def log_debug(self, message: str) -> None:
        """Log a debug message."""
        if self.controls.get_log_level() <= LOG_LEVELS["DEBUG"]:
            self.add_export_log(message, "Debug", "System")
            self.frame.panel.log_notebook.system_log_panel.log(
                message,
                "Debug",
                LOG_COLOR_DEBUG,
            )

    def log_info(self, message: str) -> None:
        """Log an informational message."""
        if self.controls.get_log_level() <= LOG_LEVELS["INFO"]:
            self.add_export_log(message, "Info", "System")
            self.frame.panel.log_notebook.system_log_panel.log(
                message,
                "Info",
                LOG_COLOR_INFO,
            )

    def log_warning(self, message: str) -> None:
        """Log a warning message."""
        if self.controls.get_log_level() <= LOG_LEVELS["WARNING"]:
            self.add_export_log(message, "Warning", "System")
            self.frame.panel.log_notebook.system_log_panel.log(
                message,
                "Warning",
                LOG_COLOR_WARNING,
            )
            self.frame.panel.log_notebook.highlight(LOG_LEVELS["WARNING"])

    def log_error(self, message: str) -> None:
        """Log an error message."""
        if self.controls.get_log_level() <= LOG_LEVELS["ERROR"]:
            self.add_export_log(message, "Error", "System")
            self.frame.panel.log_notebook.system_log_panel.log(
                message,
                "Error",
                LOG_COLOR_ERROR,
            )
            self.frame.panel.log_notebook.highlight(LOG_LEVELS["ERROR"])

    def log_critical(self, message: str) -> None:
        """Log a critical error message."""
        if self.controls.get_log_level() <= LOG_LEVELS["CRITICAL"]:
            self.add_export_log(message, "Critical", "System")
            self.frame.panel.log_notebook.system_log_panel.log(
                message,
                "Critical",
                LOG_COLOR_CRITICAL,
            )
            self.frame.panel.log_notebook.highlight(LOG_LEVELS["CRITICAL"])

    def log_context(self, message: str, silent: bool = False) -> None:
        """Log a context message."""
        tags = []
        colors = []

        if silent:
            tags.append("silent")
            colors.append(LOG_COLOR_CONTEXT_SILENT)

        self.add_export_log(message, tags, "Context")
        self.frame.panel.log_notebook.context_log_panel.log(
            message,
            tags,
            colors,
        )

    def log_description(self, message: str) -> None:
        """Log an action description."""
        self.add_export_log(message, "Action", "Context")
        self.frame.panel.log_notebook.context_log_panel.log(
            message,
            "Action",
            LOG_COLOR_CONTEXT_ACTION,
        )

    def log_query(self, message: str, ephemeral: bool = False) -> None:
        """Log an actions/force query."""
        tags = ["Query"]
        colors = [LOG_COLOR_CONTEXT_QUERY]

        if ephemeral:
            tags.append("ephemeral")
            colors.append(LOG_COLOR_CONTEXT_EPHEMERAL)

        self.add_export_log(message, tags, "Context")
        self.frame.panel.log_notebook.context_log_panel.log(
            message,
            tags,
            colors,
        )

    def log_state(self, message: str, ephemeral: bool = False) -> None:
        """Log an actions/force state."""
        tags = ["State"]
        colors = [LOG_COLOR_CONTEXT_STATE]

        if ephemeral:
            tags.append("Ephemeral")
            colors.append(LOG_COLOR_CONTEXT_EPHEMERAL)

        self.add_export_log(message, tags, "Context")
        self.frame.panel.log_notebook.context_log_panel.log(
            message,
            tags,
            colors,
        )

    def log_action_result(self, success: bool, message: str) -> None:
        """Log an action result message."""
        self.add_export_log(message, ["Result", "Success" if success else "Failure"], "Context")
        self.frame.panel.log_notebook.context_log_panel.log(
            message,
            "Result",
            LOG_COLOR_CONTEXT_ACTION_RESULT_SUCCESS if success else LOG_COLOR_CONTEXT_ACTION_RESULT_FAILURE,
        )

    def log_raw(self, message: str, incoming: bool) -> None:
        """Log raw data."""
        tag = "Game --> Tony" if incoming else "Game <-- Tony"
        color = LOG_COLOR_INCOMING if incoming else LOG_COLOR_OUTGOING

        self.add_export_log(message, tag, "Raw")
        self.frame.panel.log_notebook.raw_log_panel.log(message, tag, color)

    def clear_logs(self) -> None:
        """Clear all logs."""
        self.frame.panel.log_notebook.system_log_panel.text.Clear()
        self.frame.panel.log_notebook.command_log_panel.text.Clear()
        self.frame.panel.log_notebook.context_log_panel.text.Clear()
        self.frame.panel.log_notebook.raw_log_panel.text.Clear()
        self.model.clear_logs()

    def add_export_log(self, message: str, tags: str | list[str] | None, export_tag: str) -> None:
        """Add a log message to the export log."""
        if isinstance(tags, str):
            tags = [tags]
        tags = [dt.now().strftime("%X")] + (tags or [])

        self.model.add_log(export_tag, f"{' '.join(f'[{tag}]' for tag in tags)} {message}")

    def show_action_dialog(self, action: NeuroAction) -> str | None:
        """Show a dialog for an action. Returns the JSON string the user entered if "Send" was clicked, otherwise None."""
        self.action_dialog = ActionDialog(
            self.frame,
            self,
            action,
        )
        result = self.action_dialog.ShowModal()
        text = self.action_dialog.text.GetValue()
        self.action_dialog.Destroy()
        self.action_dialog = None

        if result == wx.ID_OK:
            assert isinstance(text, str)
            return text
        return None

    def close_action_dialog(self) -> None:
        """Close the currently opened action dialog.

        Does nothing if no dialog is open.
        Handled as if the "Cancel" button was clicked.
        """
        if self.action_dialog is not None:
            self.action_dialog.EndModal(wx.ID_CANCEL)
            self.action_dialog = None

    def add_action(self, action: NeuroAction) -> None:
        """Add an action to the list."""
        self.frame.panel.action_list.add_action(action)

    def remove_action_by_name(self, name: str) -> None:
        """Remove an action from the list by name."""
        self.frame.panel.action_list.remove_action_by_name(name)

    def enable_actions(self) -> None:
        """Enable executing actions."""
        self.frame.panel.action_list.enable_actions(True)

    def disable_actions(self) -> None:
        """Disable executing actions."""
        self.frame.panel.action_list.enable_actions(False)

    def force_actions(
        self,
        state: str,
        query: str,
        ephemeral_context: bool,
        action_names: list[str],
        retry: bool = False,
    ) -> None:
        """Show a dialog for forcing actions."""
        actions = [action for action in self.model.actions if action.name in action_names]
        actions_force_dialog = ActionsForceDialog(
            self.frame,
            self,
            state,
            query,
            ephemeral_context,
            actions,
            retry,
        )
        result = actions_force_dialog.ShowModal()
        actions_force_dialog.Destroy()

        # Executing the action has already been handled by the dialog
        if result != wx.ID_OK:
            self.log_info("Manually ignored forced action.")

    def clear_actions(self) -> None:
        """Clear the list of actions."""
        self.frame.panel.action_list.clear()

    def on_action_result(self, success: bool, message: str | None) -> None:
        """Handle an action/result message.

        Enables the execute button.
        """
        self.enable_actions()


class MainFrame(wx.Frame):  # type: ignore[misc]
    """The main frame for the Tony."""

    def __init__(self, view: TonyView) -> None:
        """Initialize MainFrame."""
        super().__init__(None, title=f"Tony v{VERSION}")

        self.view = view
        self.panel = MainPanel(self)

        self.SetSize((850, 600))


class MainPanel(wx.Panel):  # type: ignore[misc]
    """The main window for Tony."""

    def __init__(self, parent: MainFrame) -> None:
        """Initialize MainPanel."""
        super().__init__(parent)

        self.action_list = ActionList(self, True)
        right_panel = wx.Panel(self)
        self.log_notebook = LogNotebook(right_panel)
        self.control_panel = ControlPanel(right_panel)

        self.right_panel_sizer = wx.BoxSizer(wx.VERTICAL)
        self.right_panel_sizer.Add(self.log_notebook, 1, wx.EXPAND | wx.ALL, 5)
        self.right_panel_sizer.Add(self.control_panel, 0, wx.EXPAND | wx.ALL, 5)
        right_panel.SetSizer(self.right_panel_sizer)

        right_panel.SetMinClientSize(self.right_panel_sizer.GetMinSize())
        self.right_panel_sizer.Layout()

        self.sizer = wx.BoxSizer(wx.HORIZONTAL)
        self.sizer.Add(self.action_list, 1, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(right_panel, 1, wx.EXPAND)
        self.SetSizer(self.sizer)

        # Splitter settings

    def maximize_log(self) -> None:
        """Handle maximize event."""
        self.log_notebook.restore_button.Show()
        self.log_notebook.maximize_button.Hide()

        self.control_panel.Hide()
        self.right_panel_sizer.Layout()

        self.action_list.Hide()
        self.sizer.Layout()

    def restore_log(self) -> None:
        """Handle restore event."""
        self.log_notebook.restore_button.Hide()
        self.log_notebook.maximize_button.Show()

        self.control_panel.Show()
        self.right_panel_sizer.Layout()

        self.action_list.Show()
        self.sizer.Layout()


class ActionList(wx.Panel):  # type: ignore[misc]
    """The list of actions."""

    def __init__(
        self,
        parent: MainPanel | ActionsForceDialog,
        can_delete: bool,
    ) -> None:
        """Initialize ActionList panel."""
        super().__init__(parent, style=wx.BORDER_SUNKEN)

        self.can_delete = can_delete
        self.actions_enabled = True

        self.actions: list[NeuroAction] = []

        self.list = wx.ListCtrl(self, style=wx.LC_REPORT | wx.LC_SINGLE_SEL)
        self.description_text = wx.StaticText(self)
        self.description_text.Hide()
        button_panel = wx.Panel(self)
        self.execute_button = wx.Button(button_panel, label="Execute")
        self.delete_button = wx.Button(button_panel, label="Delete")
        self.delete_all_button = wx.Button(button_panel, label="Delete all")
        self.unlock_button = wx.Button(button_panel, label="Stop waiting")

        button_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_panel_sizer.Add(self.execute_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel_sizer.Add(self.delete_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel_sizer.Add(self.delete_all_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel_sizer.Add(self.unlock_button, 0, wx.EXPAND | wx.ALL, 5)
        button_panel.SetSizer(button_panel_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.list, 1, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(self.description_text, 0, wx.EXPAND | wx.ALL, 5)
        self.sizer.Add(button_panel, 0, wx.EXPAND)
        self.SetSizer(self.sizer)

        self.Bind(wx.EVT_BUTTON, self.on_execute, self.execute_button)
        self.Bind(wx.EVT_BUTTON, self.on_delete, self.delete_button)
        self.Bind(wx.EVT_BUTTON, self.on_delete_all, self.delete_all_button)
        self.Bind(wx.EVT_BUTTON, self.on_unlock, self.unlock_button)
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.on_execute, self.list)  # ListEvent is a subclass of CommandEvent
        self.Bind(wx.EVT_LIST_KEY_DOWN, self.on_key_down, self.list)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_item_selected, self.list)
        self.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.on_item_deselected, self.list)

        self.list.InsertColumn(0, "Name", width=150)
        self.list.InsertColumn(1, "Description", width=200)
        self.list.InsertColumn(2, "Schema", width=60)

        self.execute_button.SetToolTip(
            "Execute the selected action."
            " Opens a dialog to enter JSON data if the action has a schema.",
        )  # fmt: skip
        self.delete_button.SetToolTip(
            "Delete the selected action."
            " Should only be used for testing, this is not something Neuro would normally do.",
        )
        self.delete_all_button.SetToolTip(
            "Delete all actions."
            " Should only be used for testing, this is not something Neuro would normally do.",
        )  # fmt: skip
        self.unlock_button.SetToolTip("Stop waiting for the game to send an action result.")

        # Setup
        self.execute_button.Disable()
        self.delete_button.Disable()
        self.unlock_button.Disable()

        if not self.can_delete:
            self.delete_all_button.Disable()

    def add_action(self, action: NeuroAction) -> None:
        """Add an action panel to the list."""
        self.actions.append(action)

        self.list.Append(
            [
                action.name,
                action.description,
                "Yes" if action.schema is not None and action.schema != {} else "No",
            ],
        )

    def remove_action_by_name(self, name: str) -> None:
        """Remove an action panel from the list."""
        self.actions = [action for action in self.actions if action.name != name]

        index = self.list.FindItem(-1, name)
        if index != -1:
            self.list.DeleteItem(index)
        # else:
        #     self.GetTopLevelParent().view.log_error(f'Action "{name}" not found in list.')

    def clear(self) -> None:
        """Clear the list of actions."""
        self.actions.clear()
        self.list.DeleteAllItems()

    def enable_actions(self, enable: bool) -> None:
        """Enable or disable executing actions."""
        self.actions_enabled = enable
        self.execute_button.Enable(enable and self.list.GetFirstSelected() != -1)
        self.unlock_button.Enable(not enable)

    def on_execute(self, event: wx.CommandEvent) -> None:
        """Handle execute command event."""
        event.Skip()

        index = self.list.GetFirstSelected()

        if index == -1:
            return

        action = self.actions[index]

        top: MainFrame = self.GetTopLevelParent()
        sent = top.view.on_execute(action)

        if sent:
            self.GetEventHandler().ProcessEvent(ExecuteEvent(self.GetId(), action))
        top.view.log_debug(f"Sent: {sent}")

    def on_delete(self, event: wx.CommandEvent) -> None:
        """Handle delete command event."""
        event.Skip()

        index = self.list.GetFirstSelected()

        if index == -1:
            return

        action: NeuroAction = self.actions[index]

        top: MainFrame = self.GetTopLevelParent()
        top.view.on_delete_action(action.name)

    def on_delete_all(self, event: wx.CommandEvent) -> None:
        """Handle delete all command event."""
        event.Skip()

        top: MainFrame = self.GetTopLevelParent()
        top.view.on_delete_all_actions()

    def on_unlock(self, event: wx.CommandEvent) -> None:
        """Handle unlock command event."""
        event.Skip()

        top: MainFrame = self.GetTopLevelParent()
        top.view.on_unlock()

    def on_key_down(self, event: wx.ListEvent) -> None:
        """Handle key down event."""
        event.Skip()

        if event.GetKeyCode() == wx.WXK_DELETE and self.can_delete:
            self.on_delete(event)

    def on_item_selected(self, event: wx.ListEvent) -> None:
        """Handle item selected event."""
        event.Skip()

        self.execute_button.Enable(self.actions_enabled)
        self.delete_button.Enable(self.can_delete)

        self.description_text.Show()
        self.description_text.SetLabel(self.actions[event.GetIndex()].description)
        self.description_text.Wrap(self.GetClientSize().width - 10)
        self.Layout()

    def on_item_deselected(self, event: wx.ListEvent) -> None:
        """Handle item deselected event."""
        event.Skip()

        self.execute_button.Disable()
        self.delete_button.Disable()

        self.description_text.Hide()
        self.Layout()


class LogNotebook(wx.Panel):  # type: ignore[misc]
    """The notebook for logging messages."""

    def __init__(self, parent: MainPanel) -> None:
        """Initialize Log Notebook."""
        super().__init__(parent)

        self.highlight_level = 0

        # Create controls
        self.notebook = wx.Notebook(self)
        self.system_log_panel = LogPanel(self.notebook)
        self.command_log_panel = LogPanel(self.notebook)
        self.context_log_panel = LogPanel(self.notebook)
        self.raw_log_panel = LogPanel(
            self.notebook,
            text_ctrl_style=wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH | wx.HSCROLL,
        )

        button_panel = wx.Panel(self)
        self.restore_button = wx.Button(button_panel, label="Restore")
        self.maximize_button = wx.Button(button_panel, label="Maximize")
        self.clear_button = wx.Button(button_panel, label="Clear")
        self.export_button = wx.Button(button_panel, label="Export")
        self.restore_button.Hide()

        self.notebook.AddPage(self.system_log_panel, "System")
        self.notebook.AddPage(self.command_log_panel, "Commands")
        self.notebook.AddPage(self.context_log_panel, "Context")
        self.notebook.AddPage(self.raw_log_panel, "Raw")

        button_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_panel_sizer.Add(self.restore_button, 1, wx.EXPAND | wx.ALL, 2)
        button_panel_sizer.Add(self.maximize_button, 1, wx.EXPAND | wx.ALL, 2)
        button_panel_sizer.Add(self.clear_button, 1, wx.EXPAND | wx.ALL, 2)
        button_panel_sizer.Add(self.export_button, 1, wx.EXPAND | wx.ALL, 2)
        button_panel.SetSizer(button_panel_sizer)

        # Create sizer
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.notebook, 1, wx.EXPAND)
        self.sizer.Add(button_panel, 0, wx.EXPAND | wx.ALL, 0)
        self.SetSizer(self.sizer)

        # Tab icons
        image_list = wx.ImageList(16, 16)
        self.img_warning = image_list.Add(wx.ArtProvider.GetBitmap(wx.ART_WARNING, wx.ART_OTHER, (16, 16)))
        self.img_error = image_list.Add(wx.ArtProvider.GetBitmap(wx.ART_ERROR, wx.ART_OTHER, (16, 16)))
        self.notebook.AssignImageList(image_list)

        # Bind events
        self.Bind(wx.EVT_BUTTON, self.on_restore, self.restore_button)
        self.Bind(wx.EVT_BUTTON, self.on_maximize, self.maximize_button)
        self.Bind(wx.EVT_BUTTON, self.on_clear, self.clear_button)
        self.Bind(wx.EVT_BUTTON, self.on_export, self.export_button)

        self.Bind(wx.EVT_NOTEBOOK_PAGE_CHANGED, self.on_page_changed, self.notebook)

        # Tooltips
        self.restore_button.SetToolTip("Restore the log panel to its original size.")
        self.maximize_button.SetToolTip("Maximize the log panel to fill the entire window.")
        self.clear_button.SetToolTip("Clear all logs. Exported logs will also be cleared.")
        self.export_button.SetToolTip("Export logs to a file.")

    def highlight(self, level: int) -> None:
        """Highlight the log panel with a color."""
        if self.notebook.GetSelection() == 0:
            return  # Don't highlight when already on the system log

        if self.highlight_level > level:
            return  # Don't replace a highlight of higher level

        image = -1

        if LOG_LEVELS["WARNING"] <= level < LOG_LEVELS["ERROR"]:
            image = self.img_warning

        elif level >= LOG_LEVELS["ERROR"]:
            image = self.img_error

        self.highlight_level = level
        self.notebook.SetPageImage(0, image)

    def reset_highlight(self) -> None:
        """Reset the highlight of the log panel."""
        self.highlight_level = 0
        self.notebook.SetPageImage(0, -1)

    def on_page_changed(self, event: wx.BookCtrlEvent) -> None:
        """Handle page changed event."""
        event.Skip()

        index = event.GetSelection()

        if index == 0:
            self.reset_highlight()

    def on_restore(self, event: wx.CommandEvent) -> None:
        """Handle restore button event."""
        event.Skip()

        top: MainFrame = self.GetTopLevelParent()
        top.panel.restore_log()

    def on_clear(self, event: wx.CommandEvent) -> None:
        """Handle clear command event."""
        event.Skip()

        self.reset_highlight()

        top: MainFrame = self.GetTopLevelParent()
        top.view.on_clear_logs()

    def on_export(self, event: wx.CommandEvent) -> None:
        """Handle export command event."""
        event.Skip()

        with wx.FileDialog(
            self,
            "Export logs",
            wildcard="Log files (*.log)|*.log|Text files (*.txt)|*.txt|All files (*.*)|*.*",
            style=wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT,
        ) as file_dialog:
            assert isinstance(file_dialog, wx.FileDialog)

            file_dialog.SetFilename(f"tony-{dt.now().strftime('%Y-%m-%d-%H%M%S')}.log")

            if file_dialog.ShowModal() == wx.ID_CANCEL:
                return

            top: MainFrame = self.GetTopLevelParent()
            path = file_dialog.GetPath()
            with open(path, "w") as file:
                file.write(top.view.model.get_logs_formatted())

    def on_maximize(self, event: wx.CommandEvent) -> None:
        """Handle maximize command event."""
        event.Skip()

        top: MainFrame = self.GetTopLevelParent()
        top.panel.maximize_log()


class LogPanel(wx.Panel):  # type: ignore[misc]
    """The panel for logging messages."""

    def __init__(
        self,
        parent: LogNotebook,
        text_ctrl_style: int = wx.TE_MULTILINE | wx.TE_READONLY | wx.TE_RICH,
    ) -> None:
        """Initialize Log Panel."""
        super().__init__(parent, style=wx.BORDER_SUNKEN)

        self.text = wx.TextCtrl(self, style=text_ctrl_style)
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.text, 1, wx.EXPAND)
        self.SetSizer(self.sizer)

    def log(
        self,
        message: str | list[tuple[str, wx.Colour]],
        tags: str | list[str] | None = None,
        tag_colors: wx.Colour | list[wx.Colour] | None = None,
    ) -> None:
        """Log a message with optional tags and colors."""
        # Convert single tags and colors to lists
        if isinstance(tags, str):
            tags = [tags]
        if isinstance(tag_colors, wx.Colour):
            tag_colors = [tag_colors]

        # Convert None to empty lists
        tags = tags or []
        tag_colors = tag_colors or []

        # Add default color for tags without color
        tag_colors += [LOG_COLOR_DEFAULT] * (len(tags) - len(tag_colors))

        # Log timestamp
        top: MainFrame = self.GetTopLevelParent()
        fmt = "%H:%M:%S.%f" if top.view.controls.microsecond_precision else "%H:%M:%S"
        self.text.SetDefaultStyle(wx.TextAttr(LOG_COLOR_TIMESTAMP))
        self.text.AppendText(f"[{dt.now().strftime(fmt)}] ")

        # Log tags
        for tag, tag_color in zip(tags, tag_colors, strict=True):
            self.text.SetDefaultStyle(wx.TextAttr(tag_color))
            self.text.AppendText(f"[{tag}] ")

        # Log message
        if isinstance(message, str):
            self.text.SetDefaultStyle(wx.TextAttr(LOG_COLOR_DEFAULT))
            self.text.AppendText(f"{message}\n")
        else:
            for msg, color in message:
                self.text.SetDefaultStyle(wx.TextAttr(color))
                self.text.AppendText(f"{msg}")
            self.text.AppendText("\n")


class ControlPanel(wx.Panel):  # type: ignore[misc]
    """The panel for controlling the application."""

    def __init__(self, parent: MainPanel) -> None:
        """Initialize Control Panel."""
        super().__init__(parent, style=wx.BORDER_SUNKEN)

        self.view: TonyView = self.GetTopLevelParent().view

        # Create controls

        self.ignore_actions_force_checkbox = wx.CheckBox(self, label="Ignore forced actions")
        self.auto_send_checkbox = wx.CheckBox(self, label="Auto-answer")
        self.microsecond_precision_checkbox = wx.CheckBox(self, label="Log microseconds")

        latency_panel = wx.Panel(self)
        latency_text1 = wx.StaticText(latency_panel, label="L*tency:")
        self.latency_input = wx.TextCtrl(latency_panel, value="0", size=(50, -1))
        latency_text2 = wx.StaticText(latency_panel, label="ms")

        log_level_panel = wx.Panel(self)
        log_level_text = wx.StaticText(log_level_panel, label="Log level:")
        self.log_level_choice = wx.Choice(log_level_panel, choices=[s.capitalize() for s in LOG_LEVELS])

        button_panel = wx.Panel(self)
        self.send_actions_reregister_all_button = wx.Button(button_panel, label="Clear and reregister")
        self.send_shutdown_graceful_button = wx.Button(button_panel, label="Graceful shutdown")
        self.send_shutdown_graceful_cancel_button = wx.Button(button_panel, label="Cancel shutdown")
        self.send_shutdown_immediate_button = wx.Button(button_panel, label="Immediate shutdown")

        # Create sizers

        latency_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        latency_panel_sizer.Add(latency_text1, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        latency_panel_sizer.Add(self.latency_input, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        latency_panel_sizer.Add(latency_text2, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        latency_panel.SetSizer(latency_panel_sizer)

        log_lever_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        log_lever_panel_sizer.Add(log_level_text, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        log_lever_panel_sizer.Add(self.log_level_choice, 0, wx.ALL | wx.ALIGN_CENTER, 2)
        log_level_panel.SetSizer(log_lever_panel_sizer)

        button_panel_sizer = wx.WrapSizer(wx.HORIZONTAL, wx.WRAPSIZER_DEFAULT_FLAGS)
        button_panel_sizer.Add(self.send_actions_reregister_all_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.send_shutdown_graceful_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.send_shutdown_graceful_cancel_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.send_shutdown_immediate_button, 0, wx.ALL, 2)
        button_panel.SetSizer(button_panel_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.ignore_actions_force_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.auto_send_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.microsecond_precision_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(latency_panel, 0, wx.EXPAND, 0)
        self.sizer.Add(log_level_panel, 0, wx.EXPAND, 0)
        self.sizer.Add(button_panel, 0, wx.EXPAND, 0)
        self.SetSizer(self.sizer)

        wx.CallAfter(self.SendSizeEventToParent)  # For some reason the WrapSizer isn't updated unless this is called

        # Bind events

        self.Bind(wx.EVT_CHECKBOX, self.on_ignore_actions_force, self.ignore_actions_force_checkbox)
        self.Bind(wx.EVT_CHECKBOX, self.on_auto_send, self.auto_send_checkbox)
        self.Bind(wx.EVT_CHECKBOX, self.on_microsecond_precision, self.microsecond_precision_checkbox)

        self.Bind(wx.EVT_TEXT, self.on_latency, self.latency_input)

        self.Bind(wx.EVT_CHOICE, self.on_log_level, self.log_level_choice)

        self.Bind(wx.EVT_BUTTON, self.on_send_actions_reregister_all, self.send_actions_reregister_all_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_shutdown_graceful, self.send_shutdown_graceful_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_shutdown_graceful_cancel, self.send_shutdown_graceful_cancel_button)
        self.Bind(wx.EVT_BUTTON, self.on_send_shutdown_immediate, self.send_shutdown_immediate_button)

        # Set default values

        self.ignore_actions_force_checkbox.SetValue(False)
        self.auto_send_checkbox.SetValue(False)
        self.microsecond_precision_checkbox.SetValue(False)
        self.log_level_choice.SetStringSelection(self.view.controls.get_log_level_str())

        # Add tooltips

        self.ignore_actions_force_checkbox.SetToolTip("Ignore forced actions.")
        self.auto_send_checkbox.SetToolTip(
            "Automatically answer forced actions with randomly generated data (like Randy).",
        )
        self.microsecond_precision_checkbox.SetToolTip("Use microsecond precision for timestamps.")
        self.latency_input.SetToolTip(LATENCY_TOOLTIP)
        self.log_level_choice.SetToolTip(
            "Set the log level. Exported logs will still show all messages."
            "\nDebug: Usually not relevant for normal operation."
            "\nInfo: Might be useful to diagnose issues."
            "\nWarning: A command sent or received does not comply with the API specification."
            "\nError: A command sent or received is invalid and cannot be processed."
            "\nCritical: Tony will likely not be able to recover from this error.",
        )
        self.send_actions_reregister_all_button.SetToolTip(
            "Clear all actions and request reregistration from the game."
            " This is not officially part of the API specification and may not be supported by all SDKs.",
        )
        self.send_shutdown_graceful_button.SetToolTip(
            "Request a graceful shutdown from the game."
            " This is not officially part of the API specification and may not be supported by all SDKs.",
        )
        self.send_shutdown_graceful_cancel_button.SetToolTip(
            "Cancel a graceful shutdown request."
            " This is not officially part of the API specification and may not be supported by all SDKs.",
        )
        self.send_shutdown_immediate_button.SetToolTip(
            "Request an immediate shutdown from the game."
            " This is not officially part of the API specification and may not be supported by all SDKs.",
        )

    def on_ignore_actions_force(self, event: wx.CommandEvent) -> None:
        """Handle ignore_actions_force command event."""
        event.Skip()

        self.view.controls.ignore_actions_force = event.IsChecked()

    def on_auto_send(self, event: wx.CommandEvent) -> None:
        """Handle auto_send command event."""
        event.Skip()

        self.view.controls.auto_send = event.IsChecked()

    def on_microsecond_precision(self, event: wx.CommandEvent) -> None:
        """Handle microsecond_precision command event."""
        event.Skip()

        self.view.controls.microsecond_precision = event.IsChecked()

    def on_latency(self, event: wx.CommandEvent) -> None:
        """Handle latency command event."""
        event.Skip()

        try:
            latency = int(self.latency_input.GetValue())
            if latency < 0:
                raise ValueError("Latency must be non-negative.")
            if latency > 10000:
                raise ValueError("Latency must not exceed 10000 ms.")
            self.view.controls.latency = latency
            self.latency_input.SetToolTip(LATENCY_TOOLTIP)
            self.latency_input.SetBackgroundColour(wx.NullColour)  # Default color
        except ValueError as exc:
            self.latency_input.SetToolTip(str(exc))
            self.latency_input.SetBackgroundColour(UI_COLOR_ERROR)
        self.latency_input.Refresh()

    def on_log_level(self, event: wx.CommandEvent) -> None:
        """Handle log_level command event."""
        event.Skip()

        sel = self.log_level_choice.GetSelection()
        log_level: str = self.log_level_choice.GetString(sel)
        self.view.controls.set_log_level(log_level.upper())

    def on_send_actions_reregister_all(self, event: wx.CommandEvent) -> None:
        """Handle send_actions_reregister_all command event."""
        event.Skip()

        self.view.on_send_actions_reregister_all()

    def on_send_shutdown_graceful(self, event: wx.CommandEvent) -> None:
        """Handle send_shutdown_graceful command event."""
        event.Skip()

        self.view.on_send_shutdown_graceful()

    def on_send_shutdown_graceful_cancel(self, event: wx.CommandEvent) -> None:
        """Handle send_shutdown_graceful_cancel command event."""
        event.Skip()

        self.view.on_send_shutdown_graceful_cancel()

    def on_send_shutdown_immediate(self, event: wx.CommandEvent) -> None:
        """Handle send_shutdown_immediate command event."""
        event.Skip()

        self.view.on_send_shutdown_immediate()


class ActionDialog(wx.Dialog):  # type: ignore[misc]
    """Action dialog."""

    def __init__(
        self,
        parent: MainFrame,
        view: TonyView,
        action: NeuroAction,
    ) -> None:
        """Initialize Action Dialog."""
        super().__init__(parent, title=action.name, style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER)

        self.view = view
        self.action = action
        self.allow_invalid = False

        self.target_sash_ratio = 2 / 3
        self.is_error = False

        self.content_splitter = wx.SplitterWindow(self, style=wx.SP_LIVE_UPDATE)
        self.text = wx.TextCtrl(self.content_splitter, style=wx.TE_MULTILINE | wx.HSCROLL)
        self.info = wx.TextCtrl(self.content_splitter, style=wx.TE_MULTILINE | wx.HSCROLL | wx.TE_READONLY)
        self.allow_invalid_checkbox = wx.CheckBox(self, label="Don't validate")
        button_panel = wx.Panel(self)
        self.send_button = wx.Button(button_panel, label="Send")
        self.show_schema_button = wx.Button(button_panel, label="Show Schema")
        self.regenerate_button = wx.Button(button_panel, label="Regenerate")
        self.cancel_button = wx.Button(button_panel, label="Cancel")

        button_panel_sizer = wx.BoxSizer(wx.HORIZONTAL)
        button_panel_sizer.Add(self.send_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.show_schema_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.regenerate_button, 0, wx.ALL, 2)
        button_panel_sizer.Add(self.cancel_button, 0, wx.ALL, 2)
        button_panel.SetSizer(button_panel_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(self.content_splitter, 1, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.allow_invalid_checkbox, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(button_panel, 0, wx.EXPAND)
        self.SetSizer(self.sizer)

        # Bind events
        self.Bind(wx.EVT_TEXT, self.on_value_change, self.text)
        self.Bind(wx.EVT_CHECKBOX, self.on_allow_invalid, self.allow_invalid_checkbox)
        self.Bind(wx.EVT_BUTTON, self.on_send, self.send_button)
        self.Bind(wx.EVT_BUTTON, self.on_show_schema, self.show_schema_button)
        self.Bind(wx.EVT_BUTTON, self.on_regenerate, self.regenerate_button)
        self.Bind(wx.EVT_BUTTON, self.on_cancel, self.cancel_button)
        self.Bind(wx.EVT_SPLITTER_SASH_POS_CHANGED, self.on_sash_pos_changed)
        self.Bind(wx.EVT_SIZE, self.on_size)

        # Set tooltips
        self.allow_invalid_checkbox.SetToolTip("Allow sending invalid JSON data.")
        self.send_button.SetToolTip("Send the JSON data to the client.")
        self.show_schema_button.SetToolTip("Show the JSON schema of the action.")
        self.regenerate_button.SetToolTip("Generate a new random sample.")

        # Setup
        self.content_splitter.Initialize(self.text)
        self.info.Show(False)

        self.info.SetValue(json.dumps(self.action.schema, indent=2))
        self.allow_invalid_checkbox.SetValue(self.allow_invalid)

        self.faker = JSF(action.schema)
        if action.name in view.model.last_action_data:
            self.text.SetValue(view.model.last_action_data[action.name])
        else:
            self.regenerate()

        self.SetSize((600, 400))

    def regenerate(self) -> None:
        """Regenerate the JSON data."""
        sample = self.faker.generate()
        self.text.SetValue(json.dumps(sample, indent=2))

    def on_value_change(self, event: wx.CommandEvent) -> None:
        """Handle text change."""
        event.Skip()

        try:
            json_str = self.text.GetValue()
            json_cmd = json.loads(json_str)
            jsonschema.validate(json_cmd, self.action.schema)

            self.is_error = False
            self.text.SetToolTip("")
            self.text.SetBackgroundColour(wx.NullColour)
            self.Refresh()
        except Exception as exc:
            self.is_error = True
            self.text.SetToolTip(str(exc))
            self.text.SetBackgroundColour(UI_COLOR_ERROR)
            self.Refresh()

        self.send_button.Enable(self.allow_invalid or not self.is_error)

    def on_allow_invalid(self, event: wx.CommandEvent) -> None:
        """Handle allow_invalid command event."""
        event.Skip()

        self.allow_invalid = event.IsChecked()
        self.send_button.Enable(self.allow_invalid or not self.is_error)

    def on_send(self, event: wx.CommandEvent) -> None:
        """Handle send button."""
        event.Skip()

        try:
            json_str = self.text.GetValue()
            json_cmd = json.loads(json_str)
            if not self.allow_invalid:
                jsonschema.validate(json_cmd, self.action.schema)

            self.EndModal(wx.ID_OK)
            return

        except Exception as exc:
            if isinstance(exc, jsonschema.ValidationError):
                wx.MessageBox(
                    f"JSON schema validation error: {exc}",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
            elif isinstance(exc, json.JSONDecodeError):
                wx.MessageBox(
                    f"JSON decode error: {exc}",
                    "Error",
                    wx.OK | wx.ICON_ERROR,
                )
            else:
                raise exc

    def on_show_schema(self, event: wx.CommandEvent) -> None:
        """Handle show schema button."""
        event.Skip()

        self.content_splitter.SplitVertically(self.text, self.info, int(self.GetSize()[0] * self.target_sash_ratio))

    def on_cancel(self, event: wx.CommandEvent) -> None:
        """Handle cancel button."""
        event.Skip()

        self.EndModal(wx.ID_CANCEL)

    def on_regenerate(self, event: wx.CommandEvent) -> None:
        """Handle regenerate command event."""
        event.Skip()

        self.regenerate()

    def on_sash_pos_changed(self, event: wx.SplitterEvent) -> None:
        """Handle sash position changed event."""
        event.Skip()

        if self.content_splitter.IsSplit():
            self.target_sash_ratio = self.content_splitter.GetSashPosition() / self.GetSize()[0]

    def on_size(self, event: wx.SizeEvent) -> None:
        """Handle size event."""
        event.Skip()

        self.content_splitter.SetSashPosition(int(self.target_sash_ratio * self.GetSize()[0]))


class ActionsForceDialog(wx.Dialog):  # type: ignore[misc]
    """Forced Action Dialog."""

    def __init__(
        self,
        parent: MainFrame,
        view: TonyView,
        state: str,
        query: str,
        ephemeral_context: bool,
        actions: list[NeuroAction],
        retry: bool = False,
    ) -> None:
        """Initialize Forced Action Dialog."""
        title = "Forced Action" if not retry else "Retry Forced Action"
        super().__init__(
            parent,
            title=title,
            style=wx.DEFAULT_DIALOG_STYLE | wx.RESIZE_BORDER,
        )

        self.view = view
        self.state = state
        self.query = query
        self.ephemeral_context = ephemeral_context
        self.actions = actions
        self.formatted_state: str
        try:
            self.formatted_state = json.dumps(json.loads(state), indent=2)
        except (json.JSONDecodeError, TypeError):
            self.formatted_state = state

        state_panel = wx.Panel(self)
        self.state_label = wx.StaticText(state_panel, label="State:")
        self.state_text = wx.StaticText(state_panel, label="")
        self.state_textctrl = wx.TextCtrl(state_panel, style=wx.TE_MULTILINE | wx.TE_READONLY, size=(-1, 100))
        self.state_textctrl.Hide()

        query_panel = wx.Panel(self)
        self.query_label = wx.StaticText(query_panel, label="Query:")
        self.query_text = wx.StaticText(query_panel, label="")

        self.ephemeral_label = wx.StaticText(self, label=f"Ephemeral context: {ephemeral_context}")
        self.action_list = ActionList(self, False)

        state_sizer = wx.BoxSizer(wx.HORIZONTAL)
        state_sizer.Add(self.state_label, 0, wx.TOP | wx.ALL, 2)
        state_sizer.Add(self.state_text, 1, wx.EXPAND | wx.ALL, 2)
        state_sizer.Add(self.state_textctrl, 1, wx.EXPAND | wx.ALL, 2)
        state_panel.SetSizer(state_sizer)

        query_sizer = wx.BoxSizer(wx.HORIZONTAL)
        query_sizer.Add(self.query_label, 0, wx.TOP | wx.ALL, 2)
        query_sizer.Add(self.query_text, 1, wx.EXPAND | wx.ALL, 2)
        query_panel.SetSizer(query_sizer)

        self.sizer = wx.BoxSizer(wx.VERTICAL)
        self.sizer.Add(state_panel, 0, wx.EXPAND | wx.ALL, 0)
        self.sizer.Add(query_panel, 0, wx.EXPAND | wx.ALL, 0)
        self.sizer.Add(self.ephemeral_label, 0, wx.EXPAND | wx.ALL, 2)
        self.sizer.Add(self.action_list, 1, wx.EXPAND | wx.ALL, 2)
        self.SetSizer(self.sizer)

        self.sizer.Fit(self)

        # Set tooltips
        self.ephemeral_label.SetToolTip(
            "With ephemeral context, Neuro will not remember the state and query after this action.",
        )

        # Bind events
        self.Bind(EVT_EXECUTE, self.on_execute, self.action_list)

        # Setup
        for action in actions:
            self.action_list.add_action(action)

        self.action_list.list.Select(0)

        # self.state_text.SetBackgroundColour(wx.Colour(255, 255, 255))
        # self.query_text.SetBackgroundColour(wx.Colour(255, 255, 255))

        state_width = self.state_text.GetSize()[0]
        query_width = self.query_text.GetSize()[0]
        self.state_text.SetLabel(self.formatted_state or "<None>")
        self.query_text.SetLabel(query or "<None>")
        self.state_text.Wrap(state_width)
        self.query_text.Wrap(query_width)

        # If state is too large, switch to textctrl
        if self.state_text.GetSize()[1] > 100:
            self.state_text.Hide()
            self.state_textctrl.Show()
            self.state_textctrl.SetValue(self.formatted_state)

        self.Layout()
        self.sizer.Fit(self)

    def on_execute(self, event: ExecuteEvent) -> None:
        """Handle execute command event."""
        event.Skip()

        self.EndModal(wx.ID_OK)


class Controls:
    """The content of the control panel."""

    def __init__(self) -> None:
        """Initialize control panel."""
        self.ignore_actions_force: bool = False
        self.auto_send: bool = False
        self.latency: int = 0
        self.microsecond_precision: bool = False

        self.__log_level_str: str = "INFO"
        self.__log_level: int = LOG_LEVELS["INFO"]

    def set_log_level(self, log_level: str) -> None:
        """Set the log level."""
        self.__log_level_str = log_level
        self.__log_level = LOG_LEVELS[log_level]

    def get_log_level(self) -> int:
        """Get the log level."""
        return self.__log_level

    def get_log_level_str(self) -> str:
        """Get the log level as a string."""
        return self.__log_level_str



"""Model module - Keeping track of NeuroAction objects."""

from __future__ import annotations

from typing import Any, NamedTuple


class NeuroAction(NamedTuple):
    """Neuro Action Object."""

    name: str
    description: str
    schema: dict[str, Any] | None


class TonyModel:
    """Tony Model."""

    __slots__ = ("actions", "last_action_data", "logs")

    def __init__(self) -> None:
        """Initialize Tony Model."""
        self.actions: list[NeuroAction] = []
        self.logs: dict[str, str] = {}
        self.last_action_data: dict[str, str] = {}

    def __repr__(self) -> str:
        """Return representation of this model."""
        return f"{self.__class__.__name__}()"

    def add_action(self, action: NeuroAction) -> None:
        """Add an action to the list."""
        self.actions.append(action)

    def remove_action(self, action: NeuroAction) -> None:
        """Remove an action from the list."""
        self.actions.remove(action)

    def remove_action_by_name(self, name: str) -> None:
        """Remove an action from the list by name."""
        # Iterating over tuple copy or else will have
        # error from "list modified during iteration"
        for action in tuple(self.actions):
            if action.name == name:
                self.remove_action(action)

    def clear_actions(self) -> None:
        """Clear all actions from the list."""
        self.actions.clear()

    def has_action(self, name: str) -> bool:
        """Check if an action exists in the list."""
        return any(action.name == name for action in self.actions)

    def get_action_by_name(self, name: str) -> NeuroAction | None:
        """Return an action by name."""
        for action in self.actions:
            if action.name == name:
                return action
        return None

    def add_log(self, tag: str, msg: str) -> None:
        """Add a log message."""
        if tag not in self.logs:
            self.logs[tag] = msg
        else:
            self.logs[tag] += f"\n{msg}"

    def clear_logs(self) -> None:
        """Clear all logs."""
        self.logs.clear()

    def get_logs_formatted(self) -> str:
        """Return formatted log messages."""
        return "\n\n".join(f"--- {tag} ---\n\n{log}" for tag, log in self.logs.items())



from semver import Version

from neuro_api_tony import constants


def test_version_exists() -> None:
    assert isinstance(constants.VERSION, str)


def test_version_is_semantic() -> None:
    assert Version.is_valid(constants.VERSION)


def test_package_name_exists() -> None:
    assert isinstance(constants.PACKAGE_NAME, str)



"""CLI - Command Line Interface module."""

from __future__ import annotations

import sys
from getopt import GetoptError, getopt
from typing import Any, Final

import requests
import semver
import wx

from .constants import APP_NAME, PACKAGE_NAME, PYPI_API_URL, VERSION
from .controller import TonyController

HELP_MESSAGE: Final = """
Before you ask, no, I can't print this to the console.

Usage: neuro-api-tony [OPTIONS]

Options:
    -h, --help:
        Show this help message.

    -a, --addr, --address <ADDRESS>:
        The address to start the websocket server on. Default is localhost.

    --host <HOST>:
        Alias for --addr.

    -l, --log, --log-level <LOG_LEVEL>:
        The log level to use. Default is INFO. Must be one of: DEBUG, INFO,
        WARNING, ERROR, CRITICAL.

    -p, --port <PORT>:
        The port number to start the websocket server on. Default is 8000.

    -v, --version:
        Show the version of the program.
"""


def cli_run() -> None:
    """Command line interface entry point."""
    try:
        options, _ = getopt(
            sys.argv[1:],
            "ha:l:p:v",
            [
                "help",
                "addr=",
                "address=",
                "host=",
                "log=",
                "log-level=",
                "port=",
                "update",
                "version",
            ],
        )
    except GetoptError as exc:
        message(
            message=str(exc),
            caption="Invalid Option",
            style=wx.OK | wx.ICON_ERROR,
        )
        sys.exit(1)

    address = "localhost"
    port = 8000
    log_level = "INFO"
    init_message = ""

    for option, value in options:
        match option:
            case "-h" | "--help":
                message(
                    message=HELP_MESSAGE,
                    caption="Help",
                    style=wx.OK | wx.ICON_INFORMATION,
                )
                sys.exit(0)

            case "-a" | "--addr" | "--address" | "--host":
                address = value

            case "-l" | "--log" | "--log-level":
                if value.upper() not in [
                    "DEBUG",
                    "INFO",
                    "WARNING",
                    "ERROR",
                    "CRITICAL",
                ]:
                    message(
                        message="Invalid log level. Must be one of: DEBUG, INFO, WARNING, ERROR, CRITICAL.",
                        caption="Invalid Log Level",
                        style=wx.OK | wx.ICON_ERROR,
                    )
                    sys.exit(1)
                log_level = value.upper()

            case "-p" | "--port":
                port = int(value)

            case "--update":
                message(
                    message="This option is deprecated. Please update the program using git or pip.",
                    caption="Deprecated Option",
                    style=wx.OK | wx.ICON_ERROR,
                )
                sys.exit(1)

            case "-v" | "--version":
                message(
                    message=f"{APP_NAME} v{VERSION}",
                    caption="Version Information",
                    style=wx.OK | wx.ICON_INFORMATION,
                )
                sys.exit(0)

    # Check if there are updates available
    try:
        response = requests.get(PYPI_API_URL, timeout=10).json()
        remote_version = response["info"]["version"]

        if semver.compare(remote_version, VERSION) > 0:
            init_message = (
                f"An update is available. ({VERSION} -> {remote_version})\n"
                f"Depending on your installation method, pull the latest changes from GitHub or "
                f'run "pip install --upgrade {PACKAGE_NAME}" to update.'
            )

    except ConnectionError:
        init_message = "Failed to check for updates. Please check your internet connection."

    except Exception as exc:
        init_message = f"An error occurred while checking for updates:\n{exc}"

    # Start the program
    app = wx.App()
    controller = TonyController(app, log_level)
    controller.run(address, port, init_message=init_message)


def message(*args: Any, **kwargs: Any) -> None:
    """Show a message dialog."""
    app = wx.App()
    wx.MessageBox(*args, **kwargs)
    app.MainLoop()


if __name__ == "__main__":
    cli_run()



"""API Module - Handles sending and receiving data over the Neuro Websocket API.

See https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md
for more information.
"""

from __future__ import annotations

import json
import traceback
from typing import TYPE_CHECKING, Any, Final, NamedTuple, Protocol

import jsonschema
import jsonschema.exceptions
import trio
from trio_websocket import (
    ConnectionClosed,
    WebSocketConnection,
    WebSocketRequest,
    serve_websocket,
)

if TYPE_CHECKING:
    from collections.abc import Callable

    from outcome import Outcome

from .model import NeuroAction

ACTION_NAME_ALLOWED_CHARS: Final = "abcdefghijklmnopqrstuvwxyz0123456789_-"

# See https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#action
INVALID_SCHEMA_KEYS: Final = frozenset(
    {
        "$anchor",
        "$comment",
        "$defs",
        "$dynamicAnchor",
        "$dynamicRef",
        "$id",
        "$ref",
        "$schema",
        "$vocabulary",
        "additionalProperties",
        "allOf",
        "anyOf",
        "contentEncoding",
        "contentMediaType",
        "contentSchema",
        "dependentRequired",
        "dependentSchemas",
        "deprecated",
        "description",
        "else",
        "if",
        "maxProperties",
        "minProperties",
        "not",
        "oneOf",
        "patternProperties",
        "readOnly",
        "then",
        "title",
        "unevaluatedItems",
        "unevaluatedProperties",
        "writeOnly",
    },
)


class LogCommandProtocol(Protocol):
    """Protocol for `log_command`."""

    def __call__(self, command: str, incoming: bool, addition: str | None = None) -> None:
        """Signature for `log_command`.

        Parameters
        ----------
        command : str
            The command name.
        incoming : bool
            If `True`, the command was received from the client. If `False`, the command was sent to the client.
        addition : str | None
            An optional additional message. This is used to log additional information about the command, such as the
            game name for the `startup` command.

        """


class NeuroAPI:
    """NeuroAPI class."""

    def __init__(self, run_sync_soon_threadsafe: Callable[[Callable[[], object]], object]) -> None:
        """Initialize NeuroAPI.

        Parameters
        ----------
        run_sync_soon_threadsafe : Callable[[Callable[[], object]], object]
            A function that is passed to [`trio.lowlevel.start_guest_run`](https://trio.readthedocs.io/en/stable/reference-lowlevel.html#trio.lowlevel.start_guest_run) to run a function in the main thread.
            See the Trio documentation for more information.

        """
        self._message_send_channel: trio.MemorySendChannel[str] | None = None
        self._current_game = ""
        self._current_action_id: str | None = None
        self._action_forced = False
        # Tests fail if I rename this to `_run_sync_soon_threadsafe`
        self.run_sync_soon_threadsafe = run_sync_soon_threadsafe

        # Dependency injection
        # fmt: off
        self.on_startup: Callable[[StartupCommand], None] = lambda cmd: None
        """Callback that is called when a [startup](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#startup) command is received.

        Parameters
        ----------
        cmd : StartupCommand
            The received command.
        """
        self.on_context: Callable[[ContextCommand], None] = lambda cmd: None
        """Callback that is called when a [context](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#context) command is received.

        Parameters
        ----------
        cmd : ContextCommand
            The received command.
        """
        self.on_actions_register: Callable[[ActionsRegisterCommand], None] = lambda cmd: None
        """Callback that is called when an [actions/register](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#register-actions) command is received.

        Parameters
        ----------
        cmd : ActionsRegisterCommand
            The received command.
        """
        self.on_actions_unregister: Callable[[ActionsUnregisterCommand], None] = lambda cmd: None
        """Callback that is called when an [actions/unregister](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#unregister-actions) command is received.

        Parameters
        ----------
        cmd : ActionsUnregisterCommand
            The received command.
        """
        self.on_actions_force: Callable[[ActionsForceCommand], None] = lambda cmd: None
        """Callback that is called when an [actions/force](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#force-actions) command is received.

        Parameters
        ----------
        cmd : ActionsForceCommand
            The received command.
        """
        self.on_action_result: Callable[[ActionResultCommand], None] = lambda cmd: None
        """Callback that is called when an [action/result](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#action-result) command is received.

        Parameters
        ----------
        cmd : ActionResultCommand
            The received command.
        """
        self.on_shutdown_ready: Callable[[ShutdownReadyCommand], None] = lambda cmd: None
        """Callback that is called when a [shutdown/ready](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/PROPOSALS.md#shutdown-ready) command is received.

        Parameters
        ----------
        cmd : ShutdownReadyCommand
            The received command.
        """
        self.on_unknown_command: Callable[[Any], None] = lambda cmd: None
        """Callback that is called when an unknown command is received.

        Parameters
        ----------
        cmd : Any
            The received command.
        """
        self.log_command: LogCommandProtocol = lambda command, incoming, addition=None: None
        """
        log_command(command, incoming, addition)

        Logging callback that is called when a command is received.

        Parameters
        ----------
        command : str
            The command name.
        incoming : bool
            If `True`, the command was received from the client. If `False`, the command was sent to the client.
        addition : str, optional
            An optional additional message. This is used to log additional information about the command, such as the
            game name for the `startup` command.
        """
        self.log_debug: Callable[[str], None] = lambda message: None
        """Logging callback that is called when a debug message should be logged.

        Parameters
        ----------
        message : str
            The message to log.

        """
        self.log_info: Callable[[str], None] = lambda message: None
        """Logging callback that is called when an info message should be logged.

        Parameters
        ----------
        message : str
            The message to log.

        """
        self.log_warning: Callable[[str], None] = lambda message: None
        """Logging callback that is called when a warning message should be logged.

        Parameters
        ----------
        message : str
            The message to log.

        """
        self.log_error: Callable[[str], None] = lambda message: None
        """Logging callback that is called when an error message should be logged.

        Parameters
        ----------
        message : str
            The message to log.

        """
        self.log_critical: Callable[[str], None] = lambda message: None
        """Logging callback that is called when a critical error message should be logged.

        If a critical error occurs, the API instance is in an invalid state and should not be used anymore.

        Parameters
        ----------
        message : str
            The message to log.

        """
        self.log_raw: Callable[[str, bool], None] = lambda message, incoming: None
        """Logging callback that is called when any message is received or sent.

        Parameters
        ----------
        message : str
            The raw JSON data received or sent.
        incoming : bool
            If `True`, the message was received from the client. If `False`, the message was sent to the client.

        """
        self.get_delay: Callable[[], float] = lambda: 0.0
        """A function that should return the delay in seconds to wait before sending a message.

        Returns
        -------
        float
            The delay in seconds to wait before sending a message.
        """
        # fmt: on

        self._async_library_running = False
        self._async_library_root_cancel: trio.CancelScope
        self._received_loop_close_request = False

    def start(self, address: str, port: int) -> None:
        """Start hosting the websocket server with Trio in the background.

        Parameters
        ----------
        address : str
            The address to run the websocket server on.
        port : int
            The port to run the websocket server on. No other process should be using this port.

        Raises
        ------
        OSError
            If the specified port is already in use.

        """
        if self._received_loop_close_request:
            # Attempting to shut down
            self.log_critical("Something attempted to start websocket server during shutdown, ignoring.")
            return

        if self._async_library_running:
            # Already running, skip
            self.log_critical("Something attempted to start websocket server a 2nd time, ignoring.")
            return

        def done_callback(run_outcome: Outcome[None]) -> None:
            """Handle when trio run completes."""
            assert self._async_library_running, "How can stop running if not running?"
            self._async_library_running = False
            # Unwrap to make sure exceptions are printed
            run_outcome.unwrap()

        self._async_library_running = True
        self._async_library_root_cancel = trio.CancelScope()

        async def root_run() -> None:
            """Root async run, wrapped with async_library_root_cancel so it's able to be stopped remotely."""
            with self._async_library_root_cancel:
                await self._run(address, port)

        try:
            # Start the Trio guest run
            trio.lowlevel.start_guest_run(
                root_run,
                done_callback=done_callback,
                run_sync_soon_threadsafe=self.run_sync_soon_threadsafe,
                host_uses_signal_set_wakeup_fd=False,
                restrict_keyboard_interrupt_to_checkpoints=True,
                strict_exception_groups=True,
            )
        except Exception:
            # Make sure async_library_running can never be in invalid state
            # even if trio fails to launch for some reason (shouldn't happen but still)
            self._async_library_running = False
            raise

    def stop(self) -> None:
        """Stop hosting background websocket server."""
        if not self._async_library_running:
            return
        self._async_library_root_cancel.cancel()

    def on_close(self, shutdown_function: Callable[[], None]) -> None:
        """Gracefully handle application quit, cancel async run properly then call shutdown_function.

        This will stop the websocket server and then call `shutdown_function`.

        Parameters
        ----------
        shutdown_function : Callable[[], None]
            A function that will be called when the websocket server is stopped.

        """
        if self._received_loop_close_request:
            self.log_critical("Already closing, ignoring 2nd close request.")
            return

        self._received_loop_close_request = True

        # Already shut down, close
        if not self._async_library_running:
            shutdown_function()
            return

        # Tell trio run to cancel
        try:
            self.stop()
        except Exception:
            # If trigger stop somehow fails, close window
            shutdown_function()

        def shutdown_then_call() -> None:
            # If still running, reschedule this function to run again
            if self._async_library_running:
                self.run_sync_soon_threadsafe(shutdown_then_call)
            else:
                # Finally shut down, call shutdown function
                shutdown_function()

        # Schedule `shutdown_function` to be called once trio run closes
        self.run_sync_soon_threadsafe(shutdown_then_call)

    async def _run(self, address: str, port: int) -> None:
        """Server run root function."""
        self.log_info(f"Starting websocket server on ws://{address}:{port}.")
        try:
            await serve_websocket(
                self._handle_websocket_request,
                address,
                port,
                ssl_context=None,
            )
        except Exception as exc:
            self.log_critical(f"Failed to start websocket server:\n{exc}")
            raise

    @property
    def client_connected(self) -> bool:
        """Is there a client connected."""
        return self._message_send_channel is not None

    async def _handle_websocket_request(
        self,
        request: WebSocketRequest,
    ) -> None:
        """Handle websocket connection request."""
        if self.client_connected:
            # 503 is "service unavailable"
            await request.reject(
                503,
                body=b"Server does not support multiple connections at once currently",
            )
            self.log_error("Another client attempted to connect, rejecting, server does not support multiple connections at once currently")  # fmt: skip
            return

        try:
            # With blocks for send and receive to handle closing when done
            # Using message_send_channel to send websocket messages synchronously
            # Zero here means no buffer, send not allowed to happen if receive channel has
            # not read prior message waiting yet.
            self._message_send_channel, receive_channel = trio.open_memory_channel[str](0)
            with self._message_send_channel, receive_channel:
                # Accept connection
                async with await request.accept() as connection:
                    await self._handle_client_connection(
                        connection,
                        receive_channel,
                    )
        finally:
            self._message_send_channel = None

    async def _handle_client_connection(
        self,
        websocket: WebSocketConnection,
        receive_channel: trio.MemoryReceiveChannel[str],
    ) -> None:
        """Handle websocket connection lifetime."""
        try:
            async with trio.open_nursery() as nursery:
                # Start running connection read and write tasks in the background
                nursery.start_soon(
                    self._handle_consumer,
                    websocket,
                    nursery.cancel_scope,
                )
                nursery.start_soon(
                    self._handle_producer,
                    websocket,
                    receive_channel,
                )
        except trio.Cancelled:
            self.log_info("Closing current websocket connection.")

    async def _handle_consumer(
        self,
        websocket: WebSocketConnection,
        cancel_scope: trio.CancelScope,
    ) -> None:
        """Handle websocket reading head."""
        while True:
            try:
                # Read message from websocket
                message = await websocket.get_message()
            except ConnectionClosed:
                self.log_info("Websocket connection closed by client.")
                break

            try:
                json_cmd = json.loads(message)
                self.log_raw(json.dumps(json_cmd, indent=2), True)

                game = json_cmd.get("game", {})
                data = json_cmd.get("data", {})

                if game == "" or not isinstance(game, str):
                    self.log_warning("Game name is not set.")
                else:
                    # Check game name
                    if json_cmd["command"] == "startup" or json_cmd["command"] == "game/startup":
                        self._current_game = game
                    elif self._current_game != game:
                        self.log_warning("Game name does not match the current game.")
                    elif self._current_game == "":
                        self.log_warning("No startup command received.")

                # Check action result when waiting for it
                if self._current_action_id is not None and json_cmd["command"] == "actions/force":
                    self.log_warning("Received actions/force while waiting for action/result.")

                # Check if an action is already forced
                if self._action_forced and json_cmd["command"] == "actions/force":
                    self.log_warning("Received actions/force while an action is already forced.")

                # Handle the command
                match json_cmd["command"]:
                    case "startup" | "game/startup":
                        self.log_command(json_cmd["command"], True, game)

                        self._current_action_id = None

                        self.on_startup(StartupCommand(game))

                        if json_cmd["command"] == "game/startup":
                            self.log_warning('"game/startup" command is deprecated. Use "startup" instead.')

                    case "context":
                        self.log_command("context", True)
                        self.on_context(
                            ContextCommand(data["message"], data["silent"]),
                        )

                    case "actions/register":
                        names = [action["name"] for action in data["actions"]]
                        self.log_command("actions/register", True, ", ".join(names))
                        actions = []

                        # Check the actions
                        for action in data["actions"]:
                            # Check the schema
                            if "schema" in action and action["schema"] != {} and action["schema"] is not None:
                                # Neuro API does not allow boolean schemas
                                if "schema" in action and isinstance(action["schema"], bool):
                                    self.log_error(f"Boolean schemas are not allowed: {action['name']}")
                                    continue

                                # Check if the schema is valid
                                try:
                                    jsonschema.Draft7Validator.check_schema(
                                        action["schema"],
                                    )
                                except jsonschema.exceptions.SchemaError as e:
                                    self.log_error(
                                        f'Invalid schema for action "{action["name"]}": {e}',
                                    )
                                    continue

                                invalid_keys = self.check_invalid_keys_recursive(action["schema"])

                                if len(invalid_keys) > 0:
                                    self.log_warning(f"Disallowed keys in schema: {', '.join(invalid_keys)}")

                            # Check for null schema
                            if "schema" in action and action["schema"] is None:
                                self.log_warning(f"Action schema is null: {action['name']}")

                            # Check the name
                            if not isinstance(action["name"], str):
                                self.log_error(f"Action name is not a string: {action['name']}")
                                continue

                            if not all(c in ACTION_NAME_ALLOWED_CHARS for c in action["name"]):
                                self.log_warning("Action name is not a lowercase string.")

                            if action["name"] == "":
                                self.log_warning("Action name is empty.")

                            # Add the action to the list
                            actions.append(action)

                        self.on_actions_register(ActionsRegisterCommand(actions))

                    case "actions/unregister":
                        self.log_command("actions/unregister", True, ", ".join(data["action_names"]))
                        self.on_actions_unregister(ActionsUnregisterCommand(data["action_names"]))

                    case "actions/force":
                        self._action_forced = True
                        self.log_command("actions/force", True, ", ".join(data["action_names"]))
                        self.on_actions_force(
                            ActionsForceCommand(
                                data.get("state"),
                                data["query"],
                                data.get("ephemeral_context", False),
                                data["action_names"],
                            ),
                        )

                    case "action/result":
                        self.log_command("action/result", True, "success" if data["success"] else "failure")

                        # Check if an action/result was expected
                        if self._current_action_id is None:
                            self.log_warning("Unexpected action/result.")
                        # Check if the action ID matches
                        elif self._current_action_id != data["id"]:
                            self.log_warning(f'Received action ID "{data["id"]}" does not match the expected action ID "{self._current_action_id}".')  # fmt: skip

                        self.log_debug(f"Action ID: {data['id']}")

                        self._current_action_id = None
                        self.on_action_result(
                            ActionResultCommand(
                                data["success"],
                                data.get("message", None),
                            ),
                        )

                    case "shutdown/ready":
                        self.log_command("shutdown/ready", True)
                        self.log_info("shutdown/ready is not officially supported.")
                        self.on_shutdown_ready(ShutdownReadyCommand())

                    case _:
                        self.log_command(json_cmd["command"], True, "Unknown command")
                        self.log_warning(f"Unknown command: {json_cmd['command']}")
                        self.on_unknown_command(json_cmd)

            except Exception as exc:
                self.log_error(f"Error while handling message: {exc}")
                traceback.print_exception(exc)
                break
        # Cancel (stop) writing head
        cancel_scope.cancel()

    async def _handle_producer(
        self,
        websocket: WebSocketConnection,
        receive_channel: trio.MemoryReceiveChannel[str],
    ) -> None:
        """Handle websocket writing head."""
        while True:
            # Wait for messages from sending side of memory channel (queue)
            message = await receive_channel.receive()

            # Artificial latency
            # Make sure never < 0 or raises ValueError
            await trio.sleep(max(0, self.get_delay()))

            # Write message
            # If connection failure happens, will crash the read head
            # because both share a nursery, ensuring connection closes
            await websocket.send_message(message)

            try:
                self.log_raw(json.dumps(json.loads(message), indent=2), False)
            except json.JSONDecodeError:
                self.log_raw(message, False)

    def _submit_message(self, message: str) -> bool:
        """Submit a message to the send queue. Return True if client connected."""
        if not self.client_connected:
            self.log_error("No client connected!")
            return False
        # Otherwise send channel should exist
        # type checkers need help understanding `self.client_connected`
        assert self._message_send_channel is not None
        self._message_send_channel.send_nowait(message)
        return True

    def send_action(self, id_: str, name: str, data: str | None) -> bool:
        """Send an action command.

        Parameters
        ----------
        id_ : str
            An arbitrary unique string that identifies the action. This is used to match the action with the result returned by the game.

        """
        self._action_forced = False

        payload = {
            "id": id_,
            "name": name,
        }
        if data is not None:
            payload["data"] = data
        obj = {
            "command": "action",
            "data": payload,
        }

        message = json.dumps(obj)

        if not self._submit_message(message):
            return False

        self._current_action_id = id_
        self.log_command("action", False, name + (" {...}" if data else ""))
        self.log_debug(f"Action ID: {id_}")

        return True

    def send_actions_reregister_all(self) -> bool:
        """Send an [actions/reregister_all](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/PROPOSALS.md#reregister-all-actions) command.

        This signals to the game to unregister all actions and reregister them.

        Returns:
        -------
        bool
            `True` if the command was sent successfully, `False` otherwise.

        Warnings:
        --------
        This command is part of the proposed API and is not officially supported yet.
        Some SDKs may not support it.

        """
        message = json.dumps(
            {
                "command": "actions/reregister_all",
            },
        )

        if not self._submit_message(message):
            return False

        self.log_command("actions/reregister_all", False)
        self.log_info("actions/reregister_all is not officially supported.")

        return True

    def send_shutdown_graceful(self, wants_shutdown: bool) -> bool:
        """Send a [shutdown/graceful](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/PROPOSALS.md#graceful-shutdown) command.

        What this command does depends on the `wants_shutdown` parameter.

        Parameters
        ----------
        wants_shutdown : bool
            If `True`, the game should prepare for shutdown. If `False`, the game should cancel any pending graceful
            shutdown.

        Returns
        -------
        bool
            `True` if the command was sent successfully, `False` otherwise.

        Warnings
        --------
        This command is part of the proposed API and is not officially supported yet.
        Some SDKs may not support it.

        """
        message = json.dumps(
            {
                "command": "shutdown/graceful",
                "data": {
                    "wants_shutdown": wants_shutdown,
                },
            },
        )

        if not self._submit_message(message):
            return False

        self.log_command("shutdown/graceful", False, f"{wants_shutdown=}")
        self.log_info("shutdown/graceful is not officially supported.")

        return True

    def send_shutdown_immediate(self) -> bool:
        """Send a [shutdown/immediate](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/PROPOSALS.md#immediate-shutdown) command.

        This signals to the game that it will be shut down within a few seconds.

        Returns:
        -------
        bool
            `True` if the command was sent successfully, `False` otherwise.

        Warnings:
        --------
        This command is part of the proposed API and is not officially supported yet.
        Some SDKs may not support it.

        """
        message = json.dumps(
            {
                "command": "shutdown/immediate",
            },
        )

        if not self._submit_message(message):
            return False

        self.log_command("shutdown/immediate", False)
        self.log_info("shutdown/immediate is not officially supported.")

        return True

    def check_invalid_keys_recursive(
        self,
        sub_schema: dict[str, Any],
    ) -> list[str]:
        """Recursively checks for invalid keys in the schema.

        Returns a list of invalid keys that were found.
        """
        invalid_keys = []

        for key, value in sub_schema.items():
            if key in INVALID_SCHEMA_KEYS:
                invalid_keys.append(key)
            elif isinstance(value, str | int | bool):
                pass
            elif isinstance(value, dict):
                invalid_keys.extend(self.check_invalid_keys_recursive(value))
            elif isinstance(value, list):
                for item in value:
                    if isinstance(item, dict):
                        invalid_keys.extend(self.check_invalid_keys_recursive(item))
            else:
                self.log_error(f"Unhandled schema value type {type(value)!r} ({value!r})")

        return invalid_keys


class StartupCommand(NamedTuple):
    """[`startup`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#startup) command."""

    game: str
    """The name of the game."""


class ContextCommand(NamedTuple):
    """[`context`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#context) command."""

    message: str
    """The context message."""
    silent: bool
    """If `True`, Neuro should not directly react to the message."""


class ActionsRegisterCommand:
    """[`actions/register`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#register-actions) command."""

    __slots__ = ("actions",)

    def __init__(self, actions: list[dict[str, Any]]) -> None:
        """Initialize actions register command."""
        # 'schema' may be omitted, so get() is used
        self.actions = [
            NeuroAction(
                action["name"],
                action["description"],
                action.get("schema"),
            )
            for action in actions
        ]
        """The list of actions to register."""


class ActionsUnregisterCommand(NamedTuple):
    """[`actions/unregister`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#unregister-actions) command."""

    action_names: list[str]
    """The list of action names to unregister."""


class ActionsForceCommand(NamedTuple):
    """[`actions/force`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#force-actions) command."""

    state: str | None
    """The state of the action."""
    query: str
    """The query string."""
    ephemeral_context: bool
    action_names: list[str]


class ActionResultCommand(NamedTuple):
    """[`action/result`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/SPECIFICATION.md#action-result) command."""

    success: bool
    message: str | None


class ShutdownReadyCommand:
    """[`shutdown/ready`](https://github.com/VedalAI/neuro-game-sdk/blob/main/API/PROPOSALS.md#shutdown-ready) command."""

    __slots__ = ()
